"use strict";
exports.__esModule = true;
/**
 * Rich functions with helpers such as [[Function1.andThen]],
 * [[Function2.apply1]] and so on.
 *
 * We support functions of arities up to 5. For each arity, we have
 * the interface ([[Function1]], [[Function2]], ...), builders are on functions
 * on [[Function1Static]], [[Function2Static]]... accessible on constants
 * named Function1, Function2,...
 *
 * It also has for instance [[Function1Static.liftOption]], which allows
 * to use functions which are not aware of [[Option]] (for instance _.find())
 * and make them take advantage of Option, or [[Function1Static.liftEither]],
 * which allow to work with [[Either]] instead of exceptions.
 *
 * Examples:
 *
 *     const combined = Function1.of((x:number)=>x+2).andThen(x=>x*3);
 *     combined(6);
 *     => 24
 *
 *     const plus5 = Function2.of((x:number,y:number)=>x+y).apply1(5);
 *     plus5(1);
 *     => 6
 */
var Option_1 = require("./Option");
var Either_1 = require("./Either");
/**
 * This is the type of the Function0 constant, which
 * offers some helper functions to deal
 * with [[Function0]] including
 * the ability to build [[Function0]]
 * from functions using [[Function0Static.of]].
 * It also offers some builtin functions like [[Function0Static.constant]].
 */
var Function0Static = /** @class */ (function () {
    function Function0Static() {
    }
    /**
     * The constant function of one parameter:
     * will always return the value you give, no
     * matter the parameter it's given.
     */
    Function0Static.prototype.constant = function (val) {
        return exports.Function0.of(function () { return val; });
    };
    /**
     * Take a one-parameter function and lift it to become a [[Function1Static]],
     * enabling you to call [[Function1.andThen]] and other such methods on it.
     */
    Function0Static.prototype.of = function (fn) {
        var r = (function () { return fn(); });
        r.andThen = function (fn2) { return exports.Function0.of(function () { return fn2(r()); }); };
        return r;
    };
    /**
     * Take a no-parameter partial function (may return undefined or throw),
     * and lift it to return an [[Option]] instead.
     * undefined becomes a [[None]], everything else a [[Some]]
     *
     *     const randOpt = Function0.liftOption(Math.random);
     *     randOpt();
     *     => Option.of(0.49884723907769635)
     *
     *     const undef = Function0.liftOption(()=>undefined);
     *     undef();
     *     => Option.none()
     *
     *     const nl = Function0.liftOption(()=>null);
     *     nl();
     *     => Option.of(null)
     *
     *     const throws = Function0.liftOption(()=>{throw "x"});
     *     throws();
     *     => Option.none()
     *
     * Also see [[Function0Static.liftNullable]], [[OptionStatic.try_]] and
     * [[OptionStatic.tryNullable]]
     */
    Function0Static.prototype.liftOption = function (fn) {
        return exports.Function0.of(function () {
            try {
                return Option_1.Option.of(fn());
            }
            catch (_a) {
                return Option_1.Option.none();
            }
        });
    };
    /**
     * Take a no-parameter partial function (may return null, undefined or throw),
     * and lift it to return an [[Option]] instead.
     * null and undefined become a [[None]], everything else a [[Some]]
     *
     *     const randOpt = Function0.liftNullable(Math.random);
     *     randOpt();
     *     => Option.of(0.49884723907769635)
     *
     *     const undef = Function0.liftNullable(()=>undefined);
     *     undef();
     *     => Option.none()
     *
     *     const nl = Function0.liftNullable(()=>null);
     *     nl();
     *     => Option.none()
     *
     *     const throws = Function0.liftNullable(()=>{throw "x"});
     *     throws();
     *     => Option.none()
     *
     * Also see [[Function0Static.liftOption]], [[OptionStatic.try_]] and
     * [[OptionStatic.tryNullable]]
     */
    Function0Static.prototype.liftNullable = function (fn) {
        return exports.Function0.of(function () {
            try {
                return Option_1.Option.ofNullable(fn());
            }
            catch (_a) {
                return Option_1.Option.none();
            }
        });
    };
    /**
     * Take a no-parameter partial function (may return undefined or throw),
     * and lift it to return an [[Either]] instead.
     *
     * Note that unlike the [[Function0Static.liftOption]] version, if
     * the function returns undefined, the liftEither version will throw
     * (the liftOption version returns None()): if you want to do
     * pure side-effects which may throw, you're better off just using
     * javascript try blocks.
     *
     * When using typescript, to help the compiler infer the left type,
     * you can either pass a second parameter like `{} as <type>`, or
     * call with `try_<L,R>(...)`.
     *
     *     const eitherRand = Function0.liftEither(Math.random, {} as string);
     *     eitherRand();
     *     => Either.right(0.49884723907769635)
     *
     *     const undef = Function0.liftEither(() => undefined);
     *     undef();
     *     => throws
     *
     *     const throws = Function0.liftEither(() => {throw "x"});
     *     throws();
     *     => Either.left("x")
     *
     * Also see [[EitherStatic.try_]]
     */
    Function0Static.prototype.liftEither = function (fn, witness) {
        return exports.Function0.of(function () {
            try {
                var r = fn();
                if (r !== undefined) {
                    return Either_1.Either.right(r);
                }
            }
            catch (err) {
                return Either_1.Either.left(err);
            }
            throw "liftEither got undefined!";
        });
    };
    return Function0Static;
}());
exports.Function0Static = Function0Static;
/**
 * The Function1 constant allows to call the [[Function0]] "static" methods.
 */
exports.Function0 = new Function0Static();
/**
 * This is the type of the Function1 constant, which
 * offers some helper functions to deal
 * with [[Function1]] including
 * the ability to build [[Function1]]
 * from functions using [[Function1Static.of]].
 * It also offers some builtin functions like [[Function1Static.constant]].
 */
var Function1Static = /** @class */ (function () {
    function Function1Static() {
    }
    /**
     * The identity function.
     */
    Function1Static.prototype.id = function () {
        return exports.Function1.of(function (x) { return x; });
    };
    /**
     * The constant function of one parameter:
     * will always return the value you give, no
     * matter the parameter it's given.
     */
    Function1Static.prototype.constant = function (val) {
        return exports.Function1.of(function (x) { return val; });
    };
    /**
     * Take a one-parameter function and lift it to become a [[Function1Static]],
     * enabling you to call [[Function1.andThen]] and other such methods on it.
     */
    Function1Static.prototype.of = function (fn) {
        var r = (function (x) { return fn(x); });
        r.andThen = function (fn2) { return exports.Function1.of(function (x) { return fn2(r(x)); }); };
        r.compose = function (fn2) { return exports.Function1.of(function (x) { return r(fn2(x)); }); };
        return r;
    };
    /**
     * Take a one-parameter partial function (may return undefined or throw),
     * and lift it to return an [[Option]] instead.
     * undefined becomes a [[None]], everything else a [[Some]]
     *
     *     const add = Function1.liftOption((x:number)=>x+1);
     *     add(1);
     *     => Option.of(2)
     *
     *     const undef = Function1.liftOption((x:number)=>undefined);
     *     undef(1);
     *     => Option.none()
     *
     *     const nl = Function1.liftOption((x:number)=>null);
     *     nl(1);
     *     => Option.some(null)
     *
     *     const throws = Function1.liftOption((x:number)=>{throw "x"});
     *     throws(1);
     *     => Option.none()
     *
     */
    Function1Static.prototype.liftOption = function (fn) {
        return exports.Function1.of(function (x) {
            try {
                return Option_1.Option.of(fn(x));
            }
            catch (_a) {
                return Option_1.Option.none();
            }
        });
    };
    /**
     * Take a one-parameter partial function (may return undefined or throw),
     * and lift it to return an [[Option]] instead.
     * null and undefined become a [[None]], everything else a [[Some]]
     *
     *     const add = Function1.liftNullable((x:number)=>x+1);
     *     add(1);
     *     => Option.of(2)
     *
     *     const undef = Function1.liftNullable((x:number)=>undefined);
     *     undef(1);
     *     => Option.none()
     *
     *     const nl = Function1.liftNullable((x:number)=>null);
     *     nl(1);
     *     => Option.none()
     *
     *     const throws = Function1.liftNullable((x:number)=>{throw "x"});
     *     throws(1);
     *     => Option.none()
     *
     */
    Function1Static.prototype.liftNullable = function (fn) {
        return exports.Function1.of(function (x) {
            try {
                return Option_1.Option.ofNullable(fn(x));
            }
            catch (_a) {
                return Option_1.Option.none();
            }
        });
    };
    /**
     * Take a one-parameter partial function (may return undefined or throw),
     * and lift it to return an [[Either]] instead.
     *
     * Note that unlike the [[Function1Static.liftOption]] version, if
     * the function returns undefined, the liftEither version will throw
     * (the liftOption version returns None()): if you want to do
     * pure side-effects which may throw, you're better off just using
     * javascript try blocks.
     *
     * When using typescript, to help the compiler infer the left type,
     * you can either pass a second parameter like `{} as <type>`, or
     * call with `try_<L,R>(...)`.
     *
     *     const add1 = Function1.liftEither((x:number) => x+1, {} as string);
     *     add1(1);
     *     => Either.right(2)
     *
     *     const undef = Function1.liftEither((x:number) => undefined);
     *     undef(1);
     *     => throws
     *
     *     const throws = Function1.liftEither((x:number) => {throw "x"});
     *     throws(1);
     *     => Either.left("x")
     */
    Function1Static.prototype.liftEither = function (fn, witness) {
        return exports.Function1.of(function (x) {
            try {
                var r = fn(x);
                if (r !== undefined) {
                    return Either_1.Either.right(r);
                }
            }
            catch (err) {
                return Either_1.Either.left(err);
            }
            throw "liftEither got undefined!";
        });
    };
    return Function1Static;
}());
exports.Function1Static = Function1Static;
/**
 * The Function1 constant allows to call the [[Function1]] "static" methods.
 */
exports.Function1 = new Function1Static();
/**
 * This is the type of the Function2 constant, which
 * offers some helper functions to deal
 * with [[Function2]] including
 * the ability to build [[Function2]]
 * from functions using [[Function2Static.of]].
 * It also offers some builtin functions like [[Function2Static.constant]].
 */
var Function2Static = /** @class */ (function () {
    function Function2Static() {
    }
    /**
     * The constant function of two parameters:
     * will always return the value you give, no
     * matter the parameters it's given.
     */
    Function2Static.prototype.constant = function (val) {
        return exports.Function2.of(function (x, y) { return val; });
    };
    /**
     * Take a two-parameter function and lift it to become a [[Function2]],
     * enabling you to call [[Function2.andThen]] and other such methods on it.
     */
    Function2Static.prototype.of = function (fn) {
        var r = (function (x, y) { return fn(x, y); });
        r.andThen = function (fn2) { return exports.Function2.of(function (x, y) { return fn2(r(x, y)); }); };
        r.curried = function () { return exports.Function1.of(function (x) { return exports.Function1.of(function (y) { return r(x, y); }); }); };
        r.tupled = function () { return exports.Function1.of(function (pair) { return r(pair[0], pair[1]); }); };
        r.flipped = function () { return exports.Function2.of(function (x, y) { return r(y, x); }); };
        r.apply1 = function (x) { return exports.Function1.of(function (y) { return r(x, y); }); };
        return r;
    };
    /**
     * Take a two-parameter partial function (may return undefined or throw),
     * and lift it to return an [[Option]] instead.
     * undefined becomes a [[None]], everything else a [[Some]]
     *
     *     const plus = Function2.liftOption((x:number,y:number)=>x+y);
     *     plus(1,2);
     *     => Option.of(3)
     *
     *     const undef = Function2.liftOption((x:number,y:number)=>undefined);
     *     undef(1,2);
     *     => Option.none()
     *
     *     const nl = Function2.liftOption((x:number,y:number)=>null);
     *     nl(1,2);
     *     => Option.some(null)
     *
     *     const throws = Function2.liftOption((x:number,y:number)=>{throw "x"});
     *     throws(1,2);
     *     => Option.none()
     */
    Function2Static.prototype.liftOption = function (fn) {
        return exports.Function2.of(function (x, y) {
            try {
                return Option_1.Option.of(fn(x, y));
            }
            catch (_a) {
                return Option_1.Option.none();
            }
        });
    };
    /**
     * Take a two-parameter partial function (may return undefined or throw),
     * and lift it to return an [[Option]] instead.
     * null and undefined become a [[None]], everything else a [[Some]]
     *
     *     const plus = Function2.liftNullable((x:number,y:number)=>x+y);
     *     plus(1,2);
     *     => Option.of(3)
     *
     *     const undef = Function2.liftNullable((x:number,y:number)=>undefined);
     *     undef(1,2);
     *     => Option.none()
     *
     *     const nl = Function2.liftNullable((x:number,y:number)=>null);
     *     nl(1,2);
     *     => Option.none()
     *
     *     const throws = Function2.liftNullable((x:number,y:number)=>{throw "x"});
     *     throws(1,2);
     *     => Option.none()
     */
    Function2Static.prototype.liftNullable = function (fn) {
        return exports.Function2.of(function (x, y) {
            try {
                return Option_1.Option.ofNullable(fn(x, y));
            }
            catch (_a) {
                return Option_1.Option.none();
            }
        });
    };
    /**
     * Take a two-parameter partial function (may return undefined or throw),
     * and lift it to return an [[Either]] instead.
     *
     * Note that unlike the [[Function2Static.liftOption]] version, if
     * the function returns undefined, the liftEither version will throw
     * (the liftOption version returns None()): if you want to do
     * pure side-effects which may throw, you're better off just using
     * javascript try blocks.
     *
     * When using typescript, to help the compiler infer the left type,
     * you can either pass a second parameter like `{} as <type>`, or
     * call with `try_<L,R>(...)`.
     *
     *     const add = Function2.liftEither((x:number,y:number) => x+y, {} as string);
     *     add(1,2);
     *     => Either.right(3)
     *
     *     const undef = Function2.liftEither((x:number,y:number) => undefined);
     *     undef(1,2);
     *     => throws
     *
     *     const throws = Function2.liftEither((x:number,y:number) => {throw "x"});
     *     throws(1,2);
     *     => Either.left("x")
     */
    Function2Static.prototype.liftEither = function (fn, witness) {
        return exports.Function2.of(function (x, y) {
            try {
                var r = fn(x, y);
                if (r !== undefined) {
                    return Either_1.Either.right(r);
                }
            }
            catch (err) {
                return Either_1.Either.left(err);
            }
            throw "liftEither got undefined!";
        });
    };
    return Function2Static;
}());
exports.Function2Static = Function2Static;
/**
 * The Function2 constant allows to call the [[Function2]] "static" methods.
 */
exports.Function2 = new Function2Static();
/**
 * This is the type of the Function3 constant, which
 * offers some helper functions to deal
 * with [[Function3]] including
 * the ability to build [[Function3]]
 * from functions using [[Function3Static.of]].
 * It also offers some builtin functions like [[Function3Static.constant]].
 */
var Function3Static = /** @class */ (function () {
    function Function3Static() {
    }
    /**
     * The constant function of three parameters:
     * will always return the value you give, no
     * matter the parameters it's given.
     */
    Function3Static.prototype.constant = function (val) {
        return exports.Function3.of(function (x, y, z) { return val; });
    };
    /**
     * Take a three-parameter function and lift it to become a [[Function3]],
     * enabling you to call [[Function3.andThen]] and other such methods on it.
     */
    Function3Static.prototype.of = function (fn) {
        var r = (function (x, y, z) { return fn(x, y, z); });
        r.andThen = function (fn2) { return exports.Function3.of(function (x, y, z) { return fn2(r(x, y, z)); }); };
        r.curried = function () { return exports.Function1.of(function (x) { return exports.Function1.of(function (y) { return exports.Function1.of(function (z) { return r(x, y, z); }); }); }); };
        r.tupled = function () { return exports.Function1.of(function (tuple) { return r(tuple[0], tuple[1], tuple[2]); }); };
        r.flipped = function () { return exports.Function3.of(function (x, y, z) { return r(z, y, x); }); };
        r.apply1 = function (x) { return exports.Function2.of(function (y, z) { return r(x, y, z); }); };
        r.apply2 = function (x, y) { return exports.Function1.of(function (z) { return r(x, y, z); }); };
        return r;
    };
    /**
     * Take a three-parameter partial function (may return undefined or throw),
     * and lift it to return an [[Option]] instead.
     * undefined becomes a [[None]], everything else a [[Some]]
     *
     *     const add3 = Function3.liftOption(
     *         (x:number,y:number,z:number)=>x+y+z);
     *     add3(1,2,3);
     *     => Option.of(6)
     *
     *     const undef = Function3.liftOption(
     *         (x:number,y:number,z:number)=>undefined);
     *     undef(1,2,3);
     *     => Option.none()
     *
     *     const nl = Function3.liftOption(
     *         (x:number,y:number,z:number)=>null);
     *     nl(1,2,3);
     *     => Option.some(null)
     *
     *     const throws = Function3.liftOption(
     *         (x:number,y:number,z:number)=>{throw "x"});
     *     throws(1,2,3);
     *     => Option.none()
     */
    Function3Static.prototype.liftOption = function (fn) {
        return exports.Function3.of(function (x, y, z) {
            try {
                return Option_1.Option.of(fn(x, y, z));
            }
            catch (_a) {
                return Option_1.Option.none();
            }
        });
    };
    /**
     * Take a three-parameter partial function (may return undefined or throw),
     * and lift it to return an [[Option]] instead.
     * null and undefined become a [[None]], everything else a [[Some]]
     *
     *     const add3 = Function3.liftNullable(
     *         (x:number,y:number,z:number)=>x+y+z);
     *     add3(1,2,3);
     *     => Option.of(6)
     *
     *     const undef = Function3.liftNullable(
     *         (x:number,y:number,z:number)=>undefined);
     *     undef(1,2,3);
     *     => Option.none()
     *
     *     const nl = Function3.liftNullable(
     *         (x:number,y:number,z:number)=>undefined);
     *     nl(1,2,3);
     *     => Option.none()
     *
     *     const throws = Function3.liftNullable(
     *         (x:number,y:number,z:number)=>{throw "x"});
     *     throws(1,2,3);
     *     => Option.none()
     */
    Function3Static.prototype.liftNullable = function (fn) {
        return exports.Function3.of(function (x, y, z) {
            try {
                return Option_1.Option.ofNullable(fn(x, y, z));
            }
            catch (_a) {
                return Option_1.Option.none();
            }
        });
    };
    /**
     * Take a three-parameter partial function (may return undefined or throw),
     * and lift it to return an [[Either]] instead.
     *
     * Note that unlike the [[Function3Static.liftOption]] version, if
     * the function returns undefined, the liftEither version will throw
     * (the liftOption version returns None()): if you want to do
     * pure side-effects which may throw, you're better off just using
     * javascript try blocks.
     *
     * When using typescript, to help the compiler infer the left type,
     * you can either pass a second parameter like `{} as <type>`, or
     * call with `try_<L,R>(...)`.
     *
     *     const add3 = Function3.liftEither((x:number,y:number,z:number) => x+y+z, {} as string);
     *     add3(1,2,3);
     *     => Either.right(6)
     *
     *     const undef = Function3.liftEither((x:number,y:number,z:number) => undefined);
     *     undef(1,2,3);
     *     => throws
     *
     *     const throws = Function3.liftEither((x:number,y:number,z:number) => {throw "x"});
     *     throws(1,2,3);
     *     => Either.left("x")
     */
    Function3Static.prototype.liftEither = function (fn, witness) {
        return exports.Function3.of(function (x, y, z) {
            try {
                var r = fn(x, y, z);
                if (r !== undefined) {
                    return Either_1.Either.right(r);
                }
            }
            catch (err) {
                return Either_1.Either.left(err);
            }
            throw "liftEither got undefined!";
        });
    };
    return Function3Static;
}());
exports.Function3Static = Function3Static;
/**
 * The Function3 constant allows to call the [[Function3]] "static" methods.
 */
exports.Function3 = new Function3Static();
/**
 * This is the type of the Function4 constant, which
 * offers some helper functions to deal
 * with [[Function4]] including
 * the ability to build [[Function4]]
 * from functions using [[Function4Static.of]].
 * It also offers some builtin functions like [[Function4Static.constant]].
 */
var Function4Static = /** @class */ (function () {
    function Function4Static() {
    }
    /**
     * The constant function of four parameters:
     * will always return the value you give, no
     * matter the parameters it's given.
     */
    Function4Static.prototype.constant = function (val) {
        return exports.Function4.of(function (x, y, z, a) { return val; });
    };
    /**
     * Take a four-parameter function and lift it to become a [[Function4]],
     * enabling you to call [[Function4.andThen]] and other such methods on it.
     */
    Function4Static.prototype.of = function (fn) {
        var r = (function (x, y, z, a) { return fn(x, y, z, a); });
        r.andThen = function (fn2) { return exports.Function4.of(function (x, y, z, a) { return fn2(r(x, y, z, a)); }); };
        r.curried = function () { return exports.Function1.of(function (x) { return exports.Function1.of(function (y) { return exports.Function1.of(function (z) { return exports.Function1.of(function (a) { return r(x, y, z, a); }); }); }); }); };
        r.tupled = function () { return exports.Function1.of(function (tuple) { return r(tuple[0], tuple[1], tuple[2], tuple[3]); }); };
        r.flipped = function () { return exports.Function4.of(function (x, y, z, a) { return r(a, z, y, x); }); };
        r.apply1 = function (x) { return exports.Function3.of(function (y, z, a) { return r(x, y, z, a); }); };
        r.apply2 = function (x, y) { return exports.Function2.of(function (z, a) { return r(x, y, z, a); }); };
        r.apply3 = function (x, y, z) { return exports.Function1.of(function (a) { return r(x, y, z, a); }); };
        return r;
    };
    /**
     * Take a four-parameter partial function (may return undefined or throw),
     * and lift it to return an [[Option]] instead.
     * undefined becomes a [[None]], everything else a [[Some]]
     *
     *     const add4 = Function4.liftOption(
     *         (x:number,y:number,z:number,a:number)=>x+y+z+a);
     *     add4(1,2,3,4);
     *     => Option.of(10)
     *
     *     const undef = Function4.liftOption(
     *         (x:number,y:number,z:number,a:number)=>undefined);
     *     undef(1,2,3,4);
     *     => Option.none()
     *
     *     const nl = Function4.liftOption(
     *         (x:number,y:number,z:number,a:number)=>null);
     *     nl(1,2,3,4);
     *     => Option.some(null)
     *
     *     const throws = Function4.liftOption(
     *         (x:number,y:number,z:number,a:number)=>{throw "x"});
     *     throws(1,2,3,4);
     *     => Option.none()
     */
    Function4Static.prototype.liftOption = function (fn) {
        return exports.Function4.of(function (x, y, z, a) {
            try {
                return Option_1.Option.of(fn(x, y, z, a));
            }
            catch (_a) {
                return Option_1.Option.none();
            }
        });
    };
    /**
     * Take a four-parameter partial function (may return undefined or throw),
     * and lift it to return an [[Option]] instead.
     * null and undefined become a [[None]], everything else a [[Some]]
     *
     *     const add4 = Function4.liftNullable(
     *         (x:number,y:number,z:number,a:number)=>x+y+z+a);
     *     add4(1,2,3,4);
     *     => Option.of(10)
     *
     *     const undef = Function4.liftNullable(
     *         (x:number,y:number,z:number,a:number)=>undefined);
     *     undef(1,2,3,4);
     *     => Option.none()
     *
     *     const nl = Function4.liftNullable(
     *         (x:number,y:number,z:number,a:number)=>null);
     *     nl(1,2,3,4);
     *     => Option.none()
     *
     *     const throws = Function4.liftNullable(
     *         (x:number,y:number,z:number,a:number)=>{throw "x"});
     *     throws(1,2,3,4);
     *     => Option.none()
     */
    Function4Static.prototype.liftNullable = function (fn) {
        return exports.Function4.of(function (x, y, z, a) {
            try {
                return Option_1.Option.ofNullable(fn(x, y, z, a));
            }
            catch (_a) {
                return Option_1.Option.none();
            }
        });
    };
    /**
     * Take a four-parameter partial function (may return undefined or throw),
     * and lift it to return an [[Either]] instead.
     *
     * Note that unlike the [[Function4Static.liftOption]] version, if
     * the function returns undefined, the liftEither version will throw
     * (the liftOption version returns None()): if you want to do
     * pure side-effects which may throw, you're better off just using
     * javascript try blocks.
     *
     * When using typescript, to help the compiler infer the left type,
     * you can either pass a second parameter like `{} as <type>`, or
     * call with `try_<L,R>(...)`.
     *
     *     const add4 = Function4.liftEither((x:number,y:number,z:number,a:number) => x+y+z+a, {} as string);
     *     add4(1,2,3,4);
     *     => Either.right(10)
     *
     *     const undef = Function4.liftEither((x:number,y:number,z:number,a:number) => undefined);
     *     undef(1,2,3,4);
     *     => throws
     *
     *     const throws = Function4.liftEither((x:number,y:number,z:number,a:number) => {throw "x"});
     *     throws(1,2,3,4);
     *     => Either.left("x")
     */
    Function4Static.prototype.liftEither = function (fn, witness) {
        return exports.Function4.of(function (x, y, z, a) {
            try {
                var r = fn(x, y, z, a);
                if (r !== undefined) {
                    return Either_1.Either.right(r);
                }
            }
            catch (err) {
                return Either_1.Either.left(err);
            }
            throw "liftEither got undefined!";
        });
    };
    return Function4Static;
}());
exports.Function4Static = Function4Static;
;
/**
 * The Function4 constant allows to call the [[Function4]] "static" methods.
 */
exports.Function4 = new Function4Static();
/**
 * This is the type of the Function5 constant, which
 * offers some helper functions to deal
 * with [[Function5]] including
 * the ability to build [[Function5]]
 * from functions using [[Function5Static.of]].
 * It also offers some builtin functions like [[Function5Static.constant]].
 */
var Function5Static = /** @class */ (function () {
    function Function5Static() {
    }
    /**
     * The constant function of five parameters:
     * will always return the value you give, no
     * matter the parameters it's given.
     */
    Function5Static.prototype.constant = function (val) {
        return exports.Function5.of(function (x, y, z, a, b) { return val; });
    };
    /**
     * Take a five-parameter function and lift it to become a [[Function5]],
     * enabling you to call [[Function5.andThen]] and other such methods on it.
     */
    Function5Static.prototype.of = function (fn) {
        var r = (function (x, y, z, a, b) { return fn(x, y, z, a, b); });
        r.andThen = function (fn2) { return exports.Function5.of(function (x, y, z, a, b) { return fn2(r(x, y, z, a, b)); }); };
        r.curried = function () { return exports.Function1.of(function (x) { return exports.Function1.of(function (y) { return exports.Function1.of(function (z) { return exports.Function1.of(function (a) { return exports.Function1.of(function (b) { return r(x, y, z, a, b); }); }); }); }); }); };
        r.tupled = function () { return exports.Function1.of(function (tuple) { return r(tuple[0], tuple[1], tuple[2], tuple[3], tuple[4]); }); };
        r.flipped = function () { return exports.Function5.of(function (x, y, z, a, b) { return r(b, a, z, y, x); }); };
        r.apply1 = function (x) { return exports.Function4.of(function (y, z, a, b) { return r(x, y, z, a, b); }); };
        r.apply2 = function (x, y) { return exports.Function3.of(function (z, a, b) { return r(x, y, z, a, b); }); };
        r.apply3 = function (x, y, z) { return exports.Function2.of(function (a, b) { return r(x, y, z, a, b); }); };
        r.apply4 = function (x, y, z, a) { return exports.Function1.of(function (b) { return r(x, y, z, a, b); }); };
        return r;
    };
    /**
     * Take a five-parameter partial function (may return undefined or throw),
     * and lift it to return an [[Option]] instead.
     * undefined becomes a [[None]], everything else a [[Some]]
     *
     *     const add5 = Function5.liftOption(
     *         (x:number,y:number,z:number,a:number,b:number)=>x+y+z+a+b);
     *     add5(1,2,3,4,5);
     *     => Option.of(15)
     *
     *     const undef = Function5.liftOption(
     *         (x:number,y:number,z:number,a:number,b:number)=>undefined);
     *     undef(1,2,3,4,5);
     *     => Option.none()
     *
     *     const nl = Function5.liftOption(
     *         (x:number,y:number,z:number,a:number,b:number)=>null);
     *     nl(1,2,3,4,5);
     *     => Option.some(null)
     *
     *     const throws = Function5.liftOption(
     *         (x:number,y:number,z:number,a:number,b:number)=>{throw "x"});
     *     throws(1,2,3,4,5);
     *     => Option.none()
     */
    Function5Static.prototype.liftOption = function (fn) {
        return exports.Function5.of(function (x, y, z, a, b) {
            try {
                return Option_1.Option.of(fn(x, y, z, a, b));
            }
            catch (_a) {
                return Option_1.Option.none();
            }
        });
    };
    /**
     * Take a five-parameter partial function (may return undefined or throw),
     * and lift it to return an [[Option]] instead.
     * null and undefined become a [[None]], everything else a [[Some]]
     *
     *     const add5 = Function5.liftNullable(
     *         (x:number,y:number,z:number,a:number,b:number)=>x+y+z+a+b);
     *     add5(1,2,3,4,5);
     *     => Option.of(15)
     *
     *     const undef = Function5.liftNullable(
     *         (x:number,y:number,z:number,a:number,b:number)=>undefined);
     *     undef(1,2,3,4,5);
     *     => Option.none()
     *
     *     const nl = Function5.liftNullable(
     *         (x:number,y:number,z:number,a:number,b:number)=>null);
     *     nl(1,2,3,4,5);
     *     => Option.none()
     *
     *     const throws = Function5.liftNullable(
     *         (x:number,y:number,z:number,a:number,b:number)=>{throw "x"});
     *     throws(1,2,3,4,5);
     *     => Option.none()
     */
    Function5Static.prototype.liftNullable = function (fn) {
        return exports.Function5.of(function (x, y, z, a, b) {
            try {
                return Option_1.Option.ofNullable(fn(x, y, z, a, b));
            }
            catch (_a) {
                return Option_1.Option.none();
            }
        });
    };
    /**
     * Take a five-parameter partial function (may return undefined or throw),
     * and lift it to return an [[Either]] instead.
     *
     * Note that unlike the [[Function5Static.liftOption]] version, if
     * the function returns undefined, the liftEither version will throw
     * (the liftOption version returns None()): if you want to do
     * pure side-effects which may throw, you're better off just using
     * javascript try blocks.
     *
     * When using typescript, to help the compiler infer the left type,
     * you can either pass a second parameter like `{} as <type>`, or
     * call with `try_<L,R>(...)`.
     *
     *     const add5 = Function5.liftEither(
     *         (x:number,y:number,z:number,a:number,b:number) => x+y+z+a+b, {} as string);
     *     add5(1,2,3,4,5);
     *     => Either.right(15)
     *
     *     const undef = Function5.liftEither(
     *         (x:number,y:number,z:number,a:number,b:number) => undefined);
     *     undef(1,2,3,4,5);
     *     => throws
     *
     *     const throws = Function5.liftEither(
     *         (x:number,y:number,z:number,a:number,b:number) => {throw "x"});
     *     throws(1,2,3,4,5);
     *     => Either.left("x")
     */
    Function5Static.prototype.liftEither = function (fn, witness) {
        return exports.Function5.of(function (x, y, z, a, b) {
            try {
                var r = fn(x, y, z, a, b);
                if (r !== undefined) {
                    return Either_1.Either.right(r);
                }
            }
            catch (err) {
                return Either_1.Either.left(err);
            }
            throw "liftEither got undefined!";
        });
    };
    return Function5Static;
}());
exports.Function5Static = Function5Static;
/**
 * The Function5 constant allows to call the [[Function5]] "static" methods.
 */
exports.Function5 = new Function5Static();
//# sourceMappingURL=Function.js.map