"use strict";
exports.__esModule = true;
var fs_1 = require("fs");
var Vector_1 = require("../src/Vector");
var Option_1 = require("../src/Option");
var Function_1 = require("../src/Function");
var ts = require("typescript");
/**
 * In this little script we extract source code from the apidoc comments
 * all over the source of prelude.ts, generate for each source file which
 * has apidoc comments containing source code samples a special apidoc-*.ts
 * test file.
 *
 * The apidoc comments must have the source indented by 4 characters, and
 * apidoc comments can specify an expected result for an expression using
 * "=>".
 */
function getCodeSamples(commentText) {
    var closedOpen = Vector_1.Vector
        .ofIterable(commentText.split("\n"))
        .map(function (str) { return str.trim(); })
        .takeWhile(function (str) { return str.indexOf("@param") < 0; }) // don't consider anything after @param
        .foldLeft({ closed: Vector_1.Vector.empty(), open: Option_1.Option.none() }, function (sofar, cur) {
        // we want to merge consecutive lines of code. a single comment may
        // contain multiple blocks of code, separated by normal text
        // => group the blocks of code together, skip the rest.
        var isCode = cur.startsWith("*    ");
        if (!isCode && sofar.open.isSome()) {
            // this line is not code but we have an open piece of code
            // => close the open piece of code
            return { closed: sofar.closed.append(sofar.open.get()), open: Option_1.Option.none() };
        }
        if (!isCode) {
            // this line is not code, no open piece of code
            // => nothing to do
            return sofar;
        }
        var lineCode = cur.replace(/^\*    /, "");
        if (sofar.open.isSome()) {
            // this is code, and we have an open piece
            // => append to it
            return { closed: sofar.closed, open: Option_1.Option.of(sofar.open.get() + "\n" + lineCode) };
        }
        // this is code, no open piece => create a new piece
        return { closed: sofar.closed, open: Option_1.Option.of(lineCode) };
    });
    return closedOpen.closed.appendAll(closedOpen.open.toVector());
}
function getCommentsInlineCode(scanner) {
    var token = scanner.scan();
    var codeSamples = Vector_1.Vector.empty();
    var curCodeSample = Vector_1.Vector.empty();
    while (token !== ts.SyntaxKind.EndOfFileToken) {
        if (!curCodeSample.isEmpty() &&
            token === ts.SyntaxKind.Identifier) {
            codeSamples = codeSamples.appendAll(curCodeSample.map(Function_1.Function2.of(getCodeSampleInfo).apply1(scanner.getTokenText())));
            curCodeSample = Vector_1.Vector.empty();
        }
        if ((token === ts.SyntaxKind.SingleLineCommentTrivia) ||
            (token === ts.SyntaxKind.MultiLineCommentTrivia) ||
            (token === ts.SyntaxKind.JSDocComment)) {
            curCodeSample = getCodeSamples(scanner.getTokenText());
        }
        token = scanner.scan();
    }
    return codeSamples;
}
function getCodeSampleInfo(identifier, codeSample) {
    var lines = Vector_1.Vector.ofIterable(codeSample.split("\n"));
    if (!lines.last().getOrThrow().trim().startsWith("=>")) {
        return {
            identifier: identifier,
            code: codeSample
        };
    }
    return {
        identifier: identifier,
        code: lines.init().mkString("\n"),
        expectedResult: lines.last().getOrThrow().replace(/^\s*=>\s*/, "")
    };
}
function storeInVariable(code) {
    if (code.replace(/[^;]/g, "").length <= 1) {
        // only one ";" or less => single line.
        return "const xvar = " + code;
    }
    // assuming several lines
    // we rely on indentation to find out where to insert
    // the const xvar = ...
    //
    // for instance...
    //
    // first LOC
    // xcx
    // blabla( <-- insert here: last line with same indentation as first LOC
    //    ...
    //    ...);
    // =>
    var lines = Vector_1.Vector.ofIterable(code.split("\n"));
    var getIndent = function (str) { return str.replace(/[^\s].*$/, "").length; };
    var codeIndent = getIndent(lines.head().getOrThrow());
    var constExprLines = lines
        .reverse()
        .takeWhile(function (l) { return getIndent(l) > codeIndent; })
        .length() + 1;
    var _a = lines.splitAt(lines.length() - constExprLines), before = _a[0], after = _a[1];
    return before
        .append(" const xvar = ")
        .appendAll(after)
        .mkString("\n");
}
function generateTestContents(fname, sampleInfo) {
    var wrap = function (x) { return "describe(\"" + fname + " docstrings\", () => { " + x + " });"; };
    if (sampleInfo.expectedResult && sampleInfo.expectedResult.startsWith("throws")) {
        return wrap("\n            it(\"" + sampleInfo.identifier + "\", () => {\n                assert.throws(() => {\n                    " + sampleInfo.code + "\n                });\n            });\n");
    }
    else if (sampleInfo.expectedResult) {
        return wrap("\n            it(\"" + sampleInfo.identifier + "\", () => {\n                resetMathRandom();\n                " + storeInVariable(sampleInfo.code) + ";\n                assert.ok(myEq(" + sampleInfo.expectedResult + ", xvar),\n                    " + sampleInfo.expectedResult + " + \" !== \" + xvar);\n                Math.random = origMathRandom;\n            });\n");
    }
    else {
        return wrap("\n            // no result to compare the output to, just make sure this compiles\n            " + sampleInfo.code + "\n");
    }
}
function generateTestFileContents(fname, samplesInfo) {
    return "\n        import { Vector } from \"../src/Vector\";\n        import { Future } from \"../src/Future\";\n        import { LinkedList, ConsLinkedList, EmptyLinkedList } from \"../src/LinkedList\";\n        import { HashSet } from \"../src/HashSet\";\n        import { HashMap } from \"../src/HashMap\";\n        import { Stream, EmptyStream, ConsStream } from \"../src/Stream\";\n        import { Function0, Function1, Function2,\n                 Function3, Function4, Function5 } from \"../src/Function\";\n        import { Predicate } from \"../src/Predicate\";\n        import { Either, Left, Right } from \"../src/Either\";\n        import { Option, Some, None } from \"../src/Option\";\n        import { instanceOf, typeOf } from \"../src/Comparison\";\n        import * as assert from 'assert';\n\n        function myEq(a:any, b:any): boolean {\n             if (a === null && b === null) {\n                 return true;\n             }\n             if ((a === null) !== (b === null)) {\n                 return false;\n             }\n             if (a === undefined && b === undefined) {\n                 return true;\n             }\n             if ((a === undefined) !== (b === undefined)) {\n                 return false;\n             }\n             if ((<any>a).isSome && (<any>b).isSome &&\n                     (<any>a).isSome() && (<any>b).isSome()) {\n                 // workaround for Option<array> equality\n                 // useful in HashMap tests\n                 return myEq(a.getOrThrow(), b.getOrThrow());\n             }\n             if (a.toArray) {\n                 // workaround for the zip test Vector.of([1,\"a\"]) equality\n                 // real Vector returns false, compilation error otherwise\n                 return myEq(a.toArray(), b.toArray());\n             }\n             if (a.equals) {\n                 return a.equals(b);\n             }\n             if (Array.isArray(a)) {\n                  if (a.length !== b.length) {\n                       return false;\n                  }\n                  for (let i=0;i<a.length;i++) {\n                      if (!myEq(a[i], b[i])) {\n                          return false;\n                      }\n                  }\n                  return true;\n             }\n             return JSON.stringify(a) === JSON.stringify(b);\n        }\n\n        // adding date support to myEq above isn't enough because I have Option<Date>\n        // in one of the tests...\n        (<any>Date.prototype).equals = function (other: Date) { return this.getTime() === other.getTime(); }\n\n        const randomValues = [0.49884723907769635, 0.3226548779864311];\n        let randomIndex = 0;\n        const origMathRandom = Math.random;\n        const mockMathRandom = () => randomValues[(randomIndex++) % randomValues.length];\n        function resetMathRandom() {\n             randomIndex = 0;\n             Math.random = mockMathRandom;\n        }\n\n        " + samplesInfo.map(Function_1.Function2.of(generateTestContents).apply1(fname)).mkString("\n") + "\n    ";
}
function generateTestCommentsFile(filepath, fname) {
    var fileContents = fs_1.readFileSync(filepath).toString();
    var scanner = ts.createScanner(ts.ScriptTarget.ES2016, false, ts.LanguageVariant.Standard, fs_1.readFileSync(filepath).toString());
    var codeSamples = getCommentsInlineCode(scanner);
    if (codeSamples.isEmpty()) {
        // nothing to do
        return;
    }
    var outputContents = generateTestFileContents(fname, codeSamples);
    fs_1.writeFileSync("tests/apidoc-" + fname, outputContents);
}
/**
 * @hidden
 */
function generateTestComments() {
    var files = fs_1.readdirSync("./src");
    files
        .filter(function (file) { return file.endsWith(".ts"); })
        .forEach(function (file) { return generateTestCommentsFile("src/" + file, file); });
}
exports.generateTestComments = generateTestComments;
generateTestComments();
//# sourceMappingURL=Comments.js.map