"use strict";
exports.__esModule = true;
var Either_1 = require("../src/Either");
var Vector_1 = require("../src/Vector");
var TestHelpers_1 = require("./TestHelpers");
var assert = require("assert");
describe("either comparison", function () {
    it("should mark equal eithers as equal", function () {
        return assert.ok(Either_1.Either.right(5).equals(Either_1.Either.right(5)));
    });
    it("should mark different eithers as not equal", function () {
        return assert.ok(!Either_1.Either.right(5).equals(Either_1.Either.right(6)));
    });
    it("should mark left as equals to left", function () {
        return assert.ok(Either_1.Either.left("x").equals(Either_1.Either.left("x")));
    });
    it("should mark left and right as not equal", function () {
        return assert.ok(!Either_1.Either.right(5).equals(Either_1.Either.left(5)));
    });
    it("should mark left and right as not equal", function () {
        return assert.ok(!Either_1.Either.left(5).equals(Either_1.Either.right(5)));
    });
    it("should return true on contains", function () {
        return assert.ok(Either_1.Either.right(5).contains(5));
    });
    it("should return false on contains on left", function () {
        return assert.ok(!Either_1.Either.left(5).contains(5));
    });
    it("should return false on contains", function () {
        return assert.ok(!Either_1.Either.right(6).contains(5));
    });
    it("doesn't throw when given another type on equals", function () { return assert.equal(false, Either_1.Either.right(1).equals([1, 2])); });
    it("doesn't throw when given null on equals", function () { return assert.equal(false, Either_1.Either.right(1).equals(null)); });
    it("empty doesn't throw when given another type on equals", function () { return assert.equal(false, Either_1.Either.left(1).equals([1, 2])); });
    it("empty doesn't throw when given null on equals", function () { return assert.equal(false, Either_1.Either.left(1).equals(null)); });
    it("should throw when comparing eithers without true equality", function () { return assert.throws(function () { return Either_1.Either.right(Vector_1.Vector.of([1])).equals(Either_1.Either.right(Vector_1.Vector.of([1]))); }); });
    it("should fail compilation on an obviously bad equality test", function () {
        return TestHelpers_1.assertFailCompile("Either.right([1]).equals(Either.right([1]))", "is not assignable to parameter");
    });
    it("should fail compilation on an obviously bad contains test", function () {
        return TestHelpers_1.assertFailCompile("Either.right([1]).contains([1])", "is not assignable to parameter");
    });
});
describe("either transformation", function () {
    it("should transform with map", function () {
        assert.ok(Either_1.Either.right(5).equals(Either_1.Either.right(4).map(function (x) { return x + 1; })));
    });
    it("should handle null as Right", function () {
        return assert.ok(Either_1.Either.right(5).map(function (x) { return null; }).equals(Either_1.Either.right(null)));
    });
    it("should transform a Right to string properly", function () {
        return assert.equal("Right(5)", Either_1.Either.right(5).toString());
    });
    it("should transform a Left to string properly", function () {
        return assert.equal("Left(5)", Either_1.Either.left(5).toString());
    });
    it("should transform with flatMap x->y", function () {
        assert.ok(Either_1.Either.right(5).equals(Either_1.Either.right(4)
            .flatMap(function (x) { return Either_1.Either.right(x + 1); })));
    });
    it("should transform with flatMap x->left", function () {
        assert.ok(Either_1.Either.left(5).equals(Either_1.Either.right(4).flatMap(function (x) { return Either_1.Either.left(5); })));
    });
    it("should transform with flatMap left->left", function () {
        assert.ok(Either_1.Either.left(4).equals(Either_1.Either.left(4).flatMap(function (x) { return Either_1.Either.right(x + 1); })));
    });
    it("should apply bimap on the left value", function () {
        return assert.ok(Either_1.Either.left(4).equals(Either_1.Either.left(3).bimap(function (x) { return x + 1; }, function (x) { return x - 1; })));
    });
    it("should apply bimap on the right value", function () {
        return assert.ok(Either_1.Either.right(2).equals(Either_1.Either.right(3).bimap(function (x) { return x + 1; }, function (x) { return x - 1; })));
    });
    it("should apply match to a right", function () {
        return assert.equal(5, Either_1.Either.right(4).match({
            Right: function (x) { return x + 1; },
            Left: function (x) { return -1; }
        }));
    });
    it("should apply match to a left", function () {
        return assert.equal(4, Either_1.Either.left(5).match({
            Right: function (x) { return 1; },
            Left: function (x) { return x - 1; }
        }));
    });
    it("should liftA2", function () { return assert.ok(Either_1.Either.right(11).equals(Either_1.Either.liftA2(function (x, y) { return x + y; }, {})(Either_1.Either.right(5), Either_1.Either.right(6)))); });
    it("should abort liftA2 on left", function () { return assert.ok(Either_1.Either.left("bad").equals(Either_1.Either.liftA2(function (x, y) { return x + y; }, {})(Either_1.Either.right(5), Either_1.Either.left("bad")))); });
    it("should liftAp", function () {
        var fn = function (x) { return x.a + x.b + x.c; };
        var lifted = Either_1.Either.liftAp(fn, {});
        assert.ok(Either_1.Either.right(14).equals(lifted({ a: Either_1.Either.right(5), b: Either_1.Either.right(6), c: Either_1.Either.right(3) })));
    });
    it("should abort liftAp on left", function () { return assert.ok(Either_1.Either.left(2).equals(Either_1.Either.liftAp(function (x) { return x.a + x.b; })({ a: Either_1.Either.right(5), b: Either_1.Either.left(2) }))); });
    it("left should provide transform", function () { return assert.equal(6, Either_1.Either.left(5).transform(function (x) { return 6; })); });
    it("right should provide transform", function () { return assert.equal(6, Either_1.Either.right(5).transform(function (x) { return 6; })); });
    it("should return the unchanged either if it was a left on filter", function () { return assert.ok(Either_1.Either.left("bad").equals(Either_1.Either.left("bad").filter(function (x) { return x < 0; }, function (v) { return ""; }))); });
    it("should return the unchanged either if it was a right and predicate passes on filter", function () { return assert.ok(Either_1.Either.right(5).equals(Either_1.Either.right(5).filter(function (x) { return x > 0; }, function (v) { return ""; }))); });
});
describe("Either helpers", function () {
    it("should do sequence when all are right", function () {
        return assert.ok(Either_1.Either.right(Vector_1.Vector.of(1, 2, 3)).equals(Either_1.Either.sequence(Vector_1.Vector.of(Either_1.Either.right(1), Either_1.Either.right(2), Either_1.Either.right(3)))));
    });
    it("should fail sequence when some are left", function () {
        return assert.ok(Either_1.Either.left(2).equals(Either_1.Either.sequence(Vector_1.Vector.of(Either_1.Either.right(1), Either_1.Either.left(2), Either_1.Either.left(3)))));
    });
});
describe("either retrieval", function () {
    it("should return the value on Right.getOrElse", function () {
        return assert.equal(5, Either_1.Either.right(5).getOrElse(6));
    });
    it("should return the alternative on Left.getOrElse", function () {
        return assert.equal(6, Either_1.Either.left(5).getOrElse(6));
    });
    it("should return the value on Right.toVector", function () {
        return assert.deepEqual([5], Either_1.Either.right(5).toVector().toArray());
    });
    it("should return empty on Left.toVector", function () {
        return assert.deepEqual([], Either_1.Either.left(5).toVector().toArray());
    });
    it("should not throw on Right.getOrThrow", function () {
        return assert.equal(5, Either_1.Either.right(5).getOrThrow());
    });
    it("should throw on Left.getOrThrow", function () {
        return assert.throws(function () { return Either_1.Either.left(5).getOrThrow(); });
    });
    it("should throw on Left.getOrThrow with custom msg", function () {
        return assert.throws(function () { return Either_1.Either.left(5).getOrThrow("my custom msg"); }, /^my custom msg$/);
    });
    it("should offer get() if i checked for isRight", function () {
        var either = Either_1.Either.right(5);
        if (either.isRight()) {
            // what we are checking here is whether this does build
            // .get() is available only on Right
            assert.equal(5, either.get());
        }
    });
    it("should offer get() if i checked against isLeft", function () {
        var either = Either_1.Either.right(5);
        if (!either.isLeft()) {
            // what we are checking here is whether this does build
            // .get() is available only on Right
            assert.equal(5, either.get());
        }
    });
    it("should offer getLeft() if i checked for isLeft", function () {
        var either = Either_1.Either.left("5");
        if (either.isLeft()) {
            // what we are checking here is whether this does build
            // .get() is available only on Left
            assert.equal("5", either.getLeft());
        }
    });
    it("should offer getLeft() if i checked against isRight", function () {
        var either = Either_1.Either.left("5");
        if (!either.isRight()) {
            // what we are checking here is whether this does build
            // .get() is available only on Left
            assert.equal("5", either.getLeft());
        }
    });
});
//# sourceMappingURL=Either.js.map