"use strict";
exports.__esModule = true;
var HashSet_1 = require("../src/HashSet");
var Vector_1 = require("../src/Vector");
var Stream_1 = require("../src/Stream");
var Option_1 = require("../src/Option");
var Tuple2_1 = require("../src/Tuple2");
var HashMap_1 = require("../src/HashMap");
var Comparison_1 = require("../src/Comparison");
var SampleData_1 = require("./SampleData");
var TestHelpers_1 = require("./TestHelpers");
var CollectionTest = require("./Collection");
var assert = require("assert");
CollectionTest.runTests("HashSet", HashSet_1.HashSet.of, HashSet_1.HashSet.empty);
describe("hashset construction basic sanity tests", function () {
    it("should overwrite identical values", function () { return assert.ok(HashSet_1.HashSet.empty().add("test").add("test")
        .equals(HashSet_1.HashSet.empty().add("test"))); });
    it("should overwrite identical with custom types", function () { return assert.ok(HashSet_1.HashSet.empty()
        .add(new SampleData_1.MyClass("a", 1))
        .add(new SampleData_1.MyClass("a", 1))
        .add(new SampleData_1.MyClass("a", 2)).equals(HashSet_1.HashSet.empty()
        .add(new SampleData_1.MyClass("a", 1))
        .add(new SampleData_1.MyClass("a", 2)))); });
    it("should overwrite identical also when built using hashset.of", function () { return assert.equal(1, HashSet_1.HashSet.of(new SampleData_1.MyClass("a", 1)).add(new SampleData_1.MyClass("a", 1)).length()); });
    it("should support addAll on a non-empty set", function () { return assert.ok(HashSet_1.HashSet.of(1, 2, 3, 4).equals(HashSet_1.HashSet.of(1, 2).addAll([3, 4]))); });
    it("should support addAll on an empty set", function () { return assert.ok(HashSet_1.HashSet.of(1, 2, 3, 4).equals(HashSet_1.HashSet.empty().addAll([1, 2, 3, 4]))); });
    it("should fail at runtime on key without equality", function () { return assert.throws(function () {
        return HashSet_1.HashSet.empty().add({ field1: 'test', field2: -1 });
    }); });
});
describe("hashset conversions", function () {
    it("should convert to array correctly", function () {
        assert.deepEqual([1, 2, 3, 4], HashSet_1.HashSet.empty()
            .add(1).add(2).add(3).add(4).toArray().sort());
    });
    it("empty should convert to array correctly", function () {
        assert.deepEqual([], HashSet_1.HashSet.empty().toArray());
    });
    it("should be created correctly from an array", function () {
        assert.deepEqual(["a", "b", "c"], HashSet_1.HashSet.ofIterable(["a", "b", "c"]).toArray().sort());
    });
    it("should be created correctly from a spread", function () {
        assert.deepEqual(["a", "b", "c"], HashSet_1.HashSet.of("a", "b", "c").toArray().sort());
    });
    it("should be displayed in a nice format by toString", function () {
        return assert.equal("HashSet('a', 'b', 'c')", HashSet_1.HashSet.of("a", "b", "c").toString());
    });
    it("should be displayed in a nice format by toString even when nulls", function () {
        return assert.equal("HashSet(null)", HashSet_1.HashSet.of(null).toString());
    });
    it("converts to string using mkString", function () {
        return assert.equal("a|c|null", HashSet_1.HashSet.of("a", null, "c").mkString("|"));
    });
    var testSet = HashSet_1.HashSet.of(new SampleData_1.MyClass("aa", 1), new SampleData_1.MyClass("aa", 2), new SampleData_1.MyClass("ba", 1));
    it("converts to array with multiple field sorting descending on field1", function () {
        assert.deepEqual([new SampleData_1.MyClass("ba", 1), new SampleData_1.MyClass("aa", 1), new SampleData_1.MyClass("aa", 2)], testSet.toArray({ sortOn: [function (x) { return x.getField2(); }, { desc: function (x) { return x.getField1(); } }] }));
    });
    it("converts to array with multiple field sorting descending on field1", function () {
        assert.deepEqual([new SampleData_1.MyClass("aa", 1), new SampleData_1.MyClass("ba", 1), new SampleData_1.MyClass("aa", 2)], testSet.toArray({ sortOn: [function (x) { return x.getField2(); }, function (x) { return x.getField1(); }] }));
    });
    it("converts to array with multiple field sorting descending on field2", function () {
        assert.deepEqual([new SampleData_1.MyClass("aa", 2), new SampleData_1.MyClass("aa", 1), new SampleData_1.MyClass("ba", 1)], testSet.toArray({ sortOn: [{ desc: function (x) { return x.getField2(); } }, function (x) { return x.getField1(); }] }));
    });
    it("converts to array with multiple field sorting both descending", function () {
        assert.deepEqual([new SampleData_1.MyClass("aa", 2), new SampleData_1.MyClass("ba", 1), new SampleData_1.MyClass("aa", 1)], testSet.toArray({ sortOn: [{ desc: function (x) { return x.getField2(); } }, { desc: function (x) { return x.getField1(); } }] }));
    });
});
describe("hashset access", function () {
    it("should return true from contains", function () {
        assert.ok(HashSet_1.HashSet.of(1, 2, 3).contains(2));
    });
    it("should return false from contains", function () {
        assert.ok(!HashSet_1.HashSet.of(1, 2, 3).contains(4));
    });
    it("supports iterator", function () {
        var total = 0;
        var iterator = HashSet_1.HashSet.of(1, 6, 3)[Symbol.iterator]();
        var curItem = iterator.next();
        while (!curItem.done) {
            total += curItem.value;
            curItem = iterator.next();
        }
        assert.equal(10, total);
    });
    it("supports empty iterator", function () {
        var total = 0;
        var iterator = HashSet_1.HashSet.empty()[Symbol.iterator]();
        var curItem = iterator.next();
        while (!curItem.done) {
            total += curItem.value;
        }
        assert.equal(0, total);
    });
    it("supports allMatch, positive case", function () { return assert.ok(HashSet_1.HashSet.of(2, 4, 8).allMatch(function (x) { return x % 2 === 0; })); });
    it("supports allMatch, negative case", function () { return assert.ok(!HashSet_1.HashSet.of(2, 5, 8).allMatch(function (x) { return x % 2 === 0; })); });
    it("supports allMatch, empty HashSet", function () { return assert.ok(HashSet_1.HashSet.empty().allMatch(function (x) { return x % 2 === 0; })); });
    it("supports anyMatch, positive case", function () { return assert.ok(HashSet_1.HashSet.of(3, 5, 8).anyMatch(function (x) { return x % 2 === 0; })); });
    it("supports anyMatch, negative case", function () { return assert.ok(!HashSet_1.HashSet.of(3, 5, 9).anyMatch(function (x) { return x % 2 === 0; })); });
    it("supports anyMatch, empty HashSet", function () { return assert.ok(!HashSet_1.HashSet.empty().anyMatch(function (x) { return x % 2 === 0; })); });
    it("correct returns single positive case", function () { return assert.equal(5, HashSet_1.HashSet.of(5).single().getOrThrow()); });
    it("correct returns single negative case", function () { return assert.ok(HashSet_1.HashSet.of(5, 6).single().isNone()); });
    it("correct returns single empty seq", function () { return assert.ok(HashSet_1.HashSet.empty().single().isNone()); });
    it("correctly partitions also after prepend", function () { return assert.deepEqual([[1, 3, 5, 7], [2, 4, 6, 8]], HashSet_1.HashSet.of(2, 3, 4, 5, 6, 7, 8).add(1).partition(function (x) { return x % 2 !== 0; })
        .map(function (v) { return v.toVector().sortOn(function (x) { return x; }).toArray(); })); });
    it("correctly infers the more precise left type on partition in case of typeguard", function () {
        // just checking that this compiles. 'charAt' is available on strings not numbers.
        // the intersect is to make sure that partition returns me a HashSet
        // not a Collection or something less precise than HashSet.
        HashSet_1.HashSet.of(1, "test", 2, "a")
            .partition(Comparison_1.typeOf("string"))[0]
            .intersect(HashSet_1.HashSet.of("a"))
            .single().getOrThrow().charAt(0);
    });
});
describe("hashset equality", function () {
    it("doesn't throw when given another type on equals", function () { return assert.equal(false, HashSet_1.HashSet.of(1).equals([1, 2])); });
    it("doesn't throw when given null on equals", function () { return assert.equal(false, HashSet_1.HashSet.of(1).equals(null)); });
    it("should refuse elements which don't offer true equality", function () { return assert.throws(function () { return HashSet_1.HashSet.of(Vector_1.Vector.of([1])); }); });
    it("should refuse elements which don't offer true equality (neither equals nor hashcode)", function () { return assert.throws(function () { return HashSet_1.HashSet.of({ name: "" }); }); });
    it("should refuse elements which don't offer true equality (no hashcode)", function () { return assert.throws(function () { return HashSet_1.HashSet.of({ name: "", equals: function (x, y) { return true; } }); }); });
    it("should refuse elements which don't offer true equality (no equals)", function () { return assert.throws(function () { return HashSet_1.HashSet.of({ name: "", hashCode: function () { return 1; } }); }); });
    it("should refuse elements which don't offer true equality (vector)", function () { return assert.throws(function () { return HashSet_1.HashSet.empty().add(Vector_1.Vector.of(Vector_1.Vector.of([1]))); }); });
    it("should refuse elements which don't offer true equality (stream)", function () { return assert.throws(function () { return HashSet_1.HashSet.of(Stream_1.Stream.of([1])); }); });
    it("should refuse elements which don't offer true equality (option)", function () { return assert.throws(function () { return HashSet_1.HashSet.of(Option_1.Option.of([1])); }); });
    it("should refuse elements which don't offer true equality (tuple2)", function () { return assert.throws(function () { return HashSet_1.HashSet.of(Tuple2_1.Tuple2.of(1, Vector_1.Vector.of([1]))); }); });
    it("should refuse elements which don't offer true equality (hashmap)", function () { return assert.throws(function () { return HashSet_1.HashSet.of(HashMap_1.HashMap.of(["a", [1]])); }); });
    it("should fail compilation on an obviously bad key type", function () {
        return TestHelpers_1.assertFailCompile("HashSet.of([1])", "Argument of type \'number[]\' is not assignable to parameter");
    });
});
describe("hashset combinations", function () {
    it("calculates the diff well", function () { return assert.ok(HashSet_1.HashSet.of(1, 2, 4).equals(HashSet_1.HashSet.of(0, 1, 2, 3, 4).diff(HashSet_1.HashSet.of(0, 3)))); });
    it("calculates the diff from empty well", function () { return assert.ok(HashSet_1.HashSet.empty().equals(HashSet_1.HashSet.empty().diff(HashSet_1.HashSet.of(0, 3)))); });
    it("calculates the intersect well", function () { return assert.ok(HashSet_1.HashSet.of(0, 3).equals(HashSet_1.HashSet.of(0, 1, 2, 3, 4).intersect(HashSet_1.HashSet.of(0, 3)))); });
    it("calculates the intersect from empty well", function () { return assert.ok(HashSet_1.HashSet.empty().equals(HashSet_1.HashSet.empty().intersect(HashSet_1.HashSet.of(0, 3)))); });
    it("calculates removeAll well", function () { return assert.ok(HashSet_1.HashSet.of(1, 2, 4).equals(HashSet_1.HashSet.of(0, 1, 2, 3, 4).removeAll([0, 3]))); });
    it("calculates removeAll from empty well", function () { return assert.ok(HashSet_1.HashSet.empty().equals(HashSet_1.HashSet.empty().removeAll([0, 3]))); });
    it("calculates remove well", function () { return assert.ok(HashSet_1.HashSet.of(0, 1, 2, 4).equals(HashSet_1.HashSet.of(0, 1, 2, 3, 4).remove(3))); });
    it("calculates remove well event if item not present", function () { return assert.ok(HashSet_1.HashSet.of(0, 1, 2, 3, 4).equals(HashSet_1.HashSet.of(0, 1, 2, 3, 4).remove(5))); });
    it("calculates remove from empty well", function () { return assert.ok(HashSet_1.HashSet.empty().equals(HashSet_1.HashSet.empty().remove(3))); });
    it("filters correctly", function () { return assert.ok(HashSet_1.HashSet.of(2, 4).equals(HashSet_1.HashSet.of(1, 2, 3, 4, 5).filter(function (x) { return x % 2 == 0; }))); });
    it("filter upcasts if possible", function () {
        // here we only want to check that the code does compile,
        // that 'x' is correctly seen has MySubclass in the 3rd line
        HashSet_1.HashSet.of(new SampleData_1.MySubclass("a", 1, "b"))
            .filter(Comparison_1.instanceOf(SampleData_1.MySubclass))
            .filter(function (x) { return x.getField3().length > 1; });
    });
    it("keeps the custom equal/hash on filter", function () { return assert.equal(1, HashSet_1.HashSet.of(new SampleData_1.MyClass("a", 1)).filter(function (x) { return true; }).add(new SampleData_1.MyClass("a", 1)).length()); });
    it("confirms subset when correct", function () {
        return assert.ok(HashSet_1.HashSet.of(1, 2, 3).isSubsetOf(HashSet_1.HashSet.of(0, 1, 2, 3, 4)));
    });
    it("rejects subset when it should", function () {
        return assert.ok(!HashSet_1.HashSet.of(1, 2, 3, 5).isSubsetOf(HashSet_1.HashSet.of(0, 1, 2, 3, 4)));
    });
});
describe("hashset transformations", function () {
    it("map works", function () { return assert.ok(HashSet_1.HashSet.of(5, 6, 7).equals(HashSet_1.HashSet.of(1, 2, 3).map(function (x) { return x + 4; }))); });
    it("flatMap works", function () { return assert.ok(HashSet_1.HashSet.of(5, 6, 7, -5, -6, -7).equals(HashSet_1.HashSet.of(1, 2, 3).flatMap(function (x) { return HashSet_1.HashSet.of(x + 4, -x - 4); }))); });
    it("mapOption works", function () { return assert.ok(HashSet_1.HashSet.of(3, 5, 7).equals(HashSet_1.HashSet.of(1, 2, 3, 4, 5, 6).mapOption(function (x) { return x % 2 == 0 ? Option_1.Option.of(x + 1) : Option_1.Option.none(); }))); });
    it("should fold correctly", function () { return assert.equal(6, HashSet_1.HashSet.of(1, 2, 3).fold(0, function (a, b) { return a + b; })); });
    it("should foldLeft correctly", function () { return assert.equal(6, HashSet_1.HashSet.of("a", "bb", "ccc").foldLeft(0, function (soFar, item) { return soFar + item.length; })); });
    it("should foldRight correctly", function () { return assert.equal(6, HashSet_1.HashSet.of("a", "bb", "ccc").foldRight(0, function (item, soFar) { return soFar + item.length; })); });
});
//# sourceMappingURL=HashSet.js.map