"use strict";
exports.__esModule = true;
var Option_1 = require("../src/Option");
var Vector_1 = require("../src/Vector");
var Comparison_1 = require("../src/Comparison");
var SampleData_1 = require("./SampleData");
var TestHelpers_1 = require("./TestHelpers");
var assert = require("assert");
describe("option comparison", function () {
    it("should mark equal options as equal", function () {
        return assert.ok(Option_1.Option.of(5).equals(Option_1.Option.of(5)));
    });
    it("should mark different options as not equal", function () {
        return assert.ok(!Option_1.Option.of(5).equals(Option_1.Option.of(6)));
    });
    it("should mark none as equals to none", function () {
        return assert.ok(Option_1.Option.none().equals(Option_1.Option.none()));
    });
    it("should mark none and some as not equal", function () {
        return assert.ok(!Option_1.Option.of(5).equals(Option_1.Option.none()));
    });
    it("should mark none and some as not equal", function () {
        return assert.ok(!Option_1.Option.none().equals(Option_1.Option.of(5)));
    });
    it("should return true on contains", function () {
        return assert.ok(Option_1.Option.of(5).contains(5));
    });
    it("should return false on contains on none", function () {
        return assert.ok(!Option_1.Option.none().contains(5));
    });
    it("should return false on contains", function () {
        return assert.ok(!Option_1.Option.of(6).contains(5));
    });
    it("doesn't throw when given another type on equals", function () { return assert.equal(false, Option_1.Option.of(1).equals([1, 2])); });
    it("doesn't throw when given null on equals", function () { return assert.equal(false, Option_1.Option.of(1).equals(null)); });
    it("empty doesn't throw when given another type on equals", function () { return assert.equal(false, Option_1.Option.none().equals([1, 2])); });
    it("empty doesn't throw when given null on equals", function () { return assert.equal(false, Option_1.Option.none().equals(null)); });
    it("should throw when comparing options without true equality", function () { return assert.throws(function () { return Option_1.Option.of(Vector_1.Vector.of([1])).equals(Option_1.Option.of(Vector_1.Vector.of([1]))); }); });
    it("should fail compilation on an obviously bad equality test", function () {
        return TestHelpers_1.assertFailCompile("Option.of([1]).equals(Option.of([1]))", "Argument of type \'" +
            "Option<number[]>\' is not assignable to parameter");
    });
    it("should fail compilation on an obviously bad contains test", function () {
        return TestHelpers_1.assertFailCompile("Option.of([1]).contains([1])", "Argument of type \'number[]\' is not assignable to parameter");
    });
});
describe("option transformation", function () {
    it("should transform with map", function () {
        assert.ok(Option_1.Option.of(5).equals(Option_1.Option.of(4).map(function (x) { return x + 1; })));
    });
    it("should handle null as Some", function () {
        return assert.ok(Option_1.Option.of(5).map(function (x) { return null; }).equals(Option_1.Option.of(null)));
    });
    it("should transform a Some to string properly", function () {
        return assert.equal("Some(5)", Option_1.Option.of(5).toString());
    });
    it("should transform a None to string properly", function () {
        return assert.equal("None()", Option_1.Option.none().toString());
    });
    it("should transform with flatMap x->y", function () {
        assert.ok(Option_1.Option.of(5).equals(Option_1.Option.of(4).flatMap(function (x) { return Option_1.Option.of(x + 1); })));
    });
    it("should transform with flatMap x->none", function () {
        assert.ok(Option_1.Option.none().equals(Option_1.Option.of(4).flatMap(function (x) { return Option_1.Option.none(); })));
    });
    it("should transform with flatMap none->none", function () {
        assert.ok(Option_1.Option.none().equals(Option_1.Option.none().flatMap(function (x) { return Option_1.Option.of(x + 1); })));
    });
    it("should filter some->some", function () {
        return assert.ok(Option_1.Option.of(5).equals(Option_1.Option.of(5).filter(function (x) { return x > 2; })));
    });
    it("should filter some->none", function () {
        return assert.ok(Option_1.Option.of(5).filter(function (x) { return x < 2; }).isNone());
    });
    it("should filter none->none", function () {
        return assert.ok(Option_1.Option.none().filter(function (x) { return x < 2; }).isNone());
    });
    it("filter upcasts if possible", function () {
        // here we only want to check that the code does compile,
        // that 'x' is correctly seen has MySubclass in the 3rd line
        Option_1.Option.of(new SampleData_1.MySubclass("a", 1, "b"))
            .filter(Comparison_1.instanceOf(SampleData_1.MySubclass))
            .filter(function (x) { return x.getField3().length > 1; });
    });
    it("should apply match to a some", function () {
        return assert.equal(5, Option_1.Option.of(4).match({
            Some: function (x) { return x + 1; },
            None: function () { return -1; }
        }));
    });
    it("should apply match to a nome", function () {
        return assert.equal(-1, Option_1.Option.none().match({
            Some: function (x) { return x + 1; },
            None: function () { return -1; }
        }));
    });
    it("some supports transform", function () {
        return assert.equal(6, Option_1.Option.of(3).transform(function (x) { return 6; }));
    });
    it("none supports transform", function () {
        return assert.equal(6, Option_1.Option.none().transform(function (x) { return 6; }));
    });
    it("should support lists combining some & none", function () {
        var a = Option_1.Option.of(5);
        var b = Option_1.Option.none();
        var c = Vector_1.Vector.of(a, b);
    });
});
describe("Option helpers", function () {
    it("should do sequence when all are some", function () {
        return assert.ok(Option_1.Option.of(Vector_1.Vector.of(1, 2, 3)).equals(Option_1.Option.sequence(Vector_1.Vector.of(Option_1.Option.of(1), Option_1.Option.of(2), Option_1.Option.of(3)))));
    });
    it("should fail sequence when some are none", function () {
        return assert.ok(Option_1.Option.sequence(Vector_1.Vector.of(Option_1.Option.of(1), Option_1.Option.none(), Option_1.Option.of(3))).isNone());
    });
    it("should liftA2", function () { return assert.ok(Option_1.Option.of(11).equals(Option_1.Option.liftA2(function (x, y) { return x + y; })(Option_1.Option.of(5), Option_1.Option.of(6)))); });
    it("should abort liftA2 on none", function () { return assert.ok(Option_1.Option.none().equals(Option_1.Option.liftA2(function (x, y) { return x + y; })(Option_1.Option.of(5), Option_1.Option.none()))); });
    it("should liftAp", function () { return assert.ok(Option_1.Option.of(14).equals(Option_1.Option.liftAp(function (x) { return x.a + x.b + x.c; })({ a: Option_1.Option.of(5), b: Option_1.Option.of(6), c: Option_1.Option.of(3) }))); });
    it("should abort liftAp on none", function () { return assert.ok(Option_1.Option.none().equals(Option_1.Option.liftAp(function (x) { return x.a + x.b; })({ a: Option_1.Option.of(5), b: Option_1.Option.none() }))); });
    it("should support ifSome on Some", function () {
        var x = 0;
        Option_1.Option.of(5).ifSome(function (v) { return x = v; });
        assert.equal(5, x);
    });
    it("should support ifSome on None", function () {
        var x = 0;
        Option_1.Option.of(5).filter(function (x) { return x < 0; }).ifSome(function (v) { return x = v; });
        assert.equal(0, x);
    });
    it("should support ifNone on Some", function () {
        var x = 0;
        Option_1.Option.of(5).ifNone(function () { return x = 5; });
        assert.equal(0, x);
    });
    it("should support ifNone on None", function () {
        var x = 0;
        Option_1.Option.of(5).filter(function (x) { return x < 0; }).ifNone(function () { return x = 5; });
        assert.equal(5, x);
    });
});
describe("option retrieval", function () {
    it("should return the value on Some.getOrElse", function () {
        return assert.equal(5, Option_1.Option.of(5).getOrElse(6));
    });
    it("should return the alternative on None.getOrElse", function () {
        return assert.equal(6, Option_1.Option.none().getOrElse(6));
    });
    it("should return the value on Some.toVector", function () {
        return assert.deepEqual([5], Option_1.Option.of(5).toVector().toArray());
    });
    it("should return empty on None.toVector", function () {
        return assert.deepEqual([], Option_1.Option.none().toVector().toArray());
    });
    it("should not throw on Some.getOrThrow", function () {
        return assert.equal(5, Option_1.Option.of(5).getOrThrow());
    });
    it("should throw on None.getOrThrow", function () {
        return assert.throws(function () { return Option_1.Option.none().getOrThrow(); });
    });
    it("should throw on None.getOrThrow with custom msg", function () {
        return assert.throws(function () { return Option_1.Option.none().getOrThrow("my custom msg"); }, /^my custom msg$/);
    });
    it("should offer get() if i checked for isSome", function () {
        var opt = Option_1.Option.of(5);
        if (opt.isSome()) {
            // what we are checking here is whether this does build
            // .get() is available only on Some not on None
            assert.equal(5, opt.get());
        }
    });
    it("should offer get() if i eliminated isNone", function () {
        var opt = Option_1.Option.of(5);
        if (!opt.isNone()) {
            // what we are checking here is whether this does build
            // .get() is available only on Some not on None
            assert.equal(5, opt.get());
        }
    });
    it("should allow some & none in a list", function () {
        var a = Option_1.Option.of(5);
        var b = Option_1.Option.none();
        if (a.isSome() && b.isNone()) {
            Vector_1.Vector.of(a.asOption(), b);
        }
    });
});
//# sourceMappingURL=Option.js.map