"use strict";
exports.__esModule = true;
var LinkedList_1 = require("../src/LinkedList");
var Vector_1 = require("../src/Vector");
var HashMap_1 = require("../src/HashMap");
var HashSet_1 = require("../src/HashSet");
var Stream_1 = require("../src/Stream");
var Option_1 = require("../src/Option");
var Predicate_1 = require("../src/Predicate");
var Comparison_1 = require("../src/Comparison");
var SampleData_1 = require("./SampleData");
var TestHelpers_1 = require("./TestHelpers");
var CollectionTest = require("./Collection");
require("mocha-testcheck").install();
var _a = require("mocha-testcheck"), gen = _a.gen, check = _a.check;
var assert = require("assert");
/**
 * @hidden
 */
function runTests(seqName, ofIterable, of, empty, unfoldRight, nonEmptySeqName_) {
    var nonEmptySeqName = nonEmptySeqName_ || seqName;
    CollectionTest.runTests(seqName, of, empty);
    describe(seqName + " creation", function () {
        it("handles of() without parameters ok", function () { return assert.deepEqual([], of().toArray()); });
        it("creates from a JS array", function () { return assert.deepEqual(["a", "b", "c"], ofIterable(["a", "b", "c"]).toArray()); });
        it("creates from a spread", function () { return assert.deepEqual(["a", "b", "c"], of("a", "b", "c").toArray()); });
        it("creates also with nulls", function () { return assert.deepEqual([1, null, 2], of(1, null, 2).toArray()); });
        it("supports ofIterable", function () { return assert.deepEqual([1, 2, 3], ofIterable([1, 2, 3]).toArray()); });
        // need to test ofIterable with a real iterable as well
        // as an array -- vector for instance has an optimized codepath
        // for arrays so we need to give a real iterable too for full coverage.
        it("supports ofIterable from real iterable", function () { return assert.deepEqual([1, 2, 3], ofIterable(Stream_1.Stream.of(1, 2, 3)).toArray()); });
        it("supports of", function () { return assert.deepEqual([1, 2, 3], of(1, 2, 3).toArray()); });
        it("supports unfoldRight", function () { return assert.deepEqual([10, 9, 8, 7, 6, 5, 4, 3, 2, 1], unfoldRight(10, function (x) { return Option_1.Option.of(x)
            .filter(function (x) { return x !== 0; })
            .map(function (x) { return [x, x - 1]; }); }).toArray()); });
    });
    describe(seqName + " prepend", function () {
        var basic = of(1, 2, 3, 4);
        var prepended = of(2, 3, 4).prepend(1);
        it("prepends correctly", function () { return assert.ok(basic.equals(prepended)); });
        it("converts to array correctly", function () { return assert.deepEqual(basic.toArray(), prepended.toArray()); });
        it("appends correctly after prepend", function () { return assert.ok(basic.append(5).equals(prepended.append(5))); });
        it("appendsAll correctly after prepend", function () { return assert.ok(basic.appendAll([5, 6]).equals(prepended.appendAll([5, 6]))); });
        it("converts to string correctly after prepend", function () { return assert.equal(basic.toString(), prepended.toString()); });
        it("prependsAll correctly", function () { return assert.deepEqual([1, 2, 3, 4, 5], of(4, 5).prependAll([1, 2, 3]).toArray()); });
    });
    describe(seqName + " Value tests", function () {
        it("has non-obviously-broken equals", function () { return assert.ok(of("a", "b", "c").equals(of("a", "b", "c"))); });
        it("doesn't throw when given another type on equals", function () { return assert.equal(false, of(1).equals([1, 2])); });
        it("doesn't throw when given null on equals", function () { return assert.equal(false, of(1).equals(null)); });
        it("is strict with equality", function () { return assert.ok(!of(1, 2).equals(of(1, undefined))); });
        it("supports contain", function () { return assert.ok(of(1, 2, 3).contains(2)); });
        it("should throw when using contains without true equality", function () { return assert.throws(function () { return of(Option_1.Option.of([1])).contains(Option_1.Option.of([1])); }); });
        it("rejects contain", function () { return assert.ok(!of(1, 2, 3).contains(4)); });
        it("rejects contain, empty stream", function () { return assert.ok(!empty().contains(4)); });
        it("supports contains, custom equality", function () { return assert.ok(of(new SampleData_1.MyClass("hi", 3)).contains(new SampleData_1.MyClass("hi", 3))); });
        it("supports allMatch, positive case", function () { return assert.ok(of(2, 4, 8).allMatch(function (x) { return x % 2 === 0; })); });
        it("supports allMatch, negative case", function () { return assert.ok(!of(2, 5, 8).allMatch(function (x) { return x % 2 === 0; })); });
        it("supports allMatch, empty stream", function () { return assert.ok(empty().allMatch(function (x) { return x % 2 === 0; })); });
        it("supports anyMatch, positive case", function () { return assert.ok(of(3, 5, 8).anyMatch(function (x) { return x % 2 === 0; })); });
        it("supports anyMatch, negative case", function () { return assert.ok(!of(3, 5, 9).anyMatch(function (x) { return x % 2 === 0; })); });
        it("supports anyMatch, empty stream", function () { return assert.ok(!empty().anyMatch(function (x) { return x % 2 === 0; })); });
        it("should fail compilation on an obviously bad equality test", function () {
            return TestHelpers_1.assertFailCompile(seqName + ".of([1]).equals(" + seqName + ".of([1]))", "Argument of type \'" +
                nonEmptySeqName + "<number[]>\' is not assignable to parameter");
        });
        it("should fail compilation on an obviously bad contains test", function () {
            return TestHelpers_1.assertFailCompile(seqName + ".of([1]).contains([1])", "Argument of type \'number[]\' is not assignable to parameter");
        });
    });
    describe(seqName + " iteration", function () {
        // can get for..of in tests by changing the target to es6,
        // or enabling downlevelIteration in the tsconfig.json,
        // not doing that for now.
        // it("supports for of", () => {
        //     let total = 0;
        //     for (const x of of(1,2,3)) {
        //         total += x;
        //     }
        //     assert.equal(6, total);
        // })
        it("finds items", function () {
            return of(1, 2, 3).find(function (x) { return x >= 2; }).contains(2);
        });
        it("doesn't find if the predicate doesn't match", function () {
            return of(1, 2, 3).find(function (x) { return x >= 4; }).isNone();
        });
        it("folds correctly", function () { return assert.equal(6, of(1, 2, 3).fold(0, function (a, b) { return a + b; })); });
        it("foldsLeft correctly", function () { return assert.equal("cba!", of("a", "b", "c").foldLeft("!", function (xs, x) { return x + xs; })); });
        it("foldsRight correctly", function () { return assert.equal("!cba", of("a", "b", "c").foldRight("!", function (x, xs) { return xs + x; })); });
        it("handles scanLeft correctly on an empty seq", function () { return assert.deepEqual([0], empty().scanLeft(0, function (i, j) { return i + j; }).toArray()); });
        it("handles scanRight correctly on an empty seq", function () { return assert.deepEqual([0], empty().scanRight(0, function (j, i) { return i + j; }).toArray()); });
        it("calls forEach correctly", function () {
            var ar = [];
            of(1, 2, 3).forEach(function (v) { return ar.push(v); });
            assert.deepEqual([1, 2, 3], ar);
        });
        it("supports iterator", function () {
            var total = 0;
            var iterator = of(1, 2, 3)[Symbol.iterator]();
            var curItem = iterator.next();
            while (!curItem.done) {
                total += curItem.value;
                curItem = iterator.next();
            }
            assert.equal(6, total);
        });
    });
    describe(seqName + " conversions", function () {
        it("mkString ok for null too", function () { return assert.equal("null", of(null).mkString(", ")); });
        it("mkString works", function () { return assert.equal("1, 2, 3", of(1, 2, 3).mkString(", ")); });
        it("mkString works for strings too", function () { return assert.equal("a, b, c", of("a", "b", "c").mkString(", ")); });
        var nullSeq = of(null);
        nullSeq.last(); // needed to force the stream to get loaded
        it("toString ok for null too", function () { return assert.equal(seqName + "(null)", nullSeq.toString()); });
        var onetwothree = of(1, 2, 3);
        onetwothree.last(); // needed to force the stream to get loaded
        it("toString works", function () { return assert.equal(seqName + "(1, 2, 3)", onetwothree.toString()); });
        var abc = of("a", "b", "c");
        abc.last(); // needed to force the stream to get loaded
        it("toString works for strings too", function () { return assert.equal(seqName + "('a', 'b', 'c')", abc.toString()); });
        it("transforms to map", function () {
            assert.ok(HashMap_1.HashMap.empty().put(1, "ok").put(2, "bad")
                .equals(of([1, "ok"], [2, "bad"]).toMap(function (x) { return x; })));
        });
        it("empty seq transforms to empty set", function () {
            assert.ok(HashSet_1.HashSet.empty()
                .equals(empty().toSet(function (x) { return x; })));
        });
    });
    describe(seqName + " manipulation", function () {
        it("computes the length correctly", function () { return assert.equal(3, of(1, 2, 3).length()); });
        it("computes the length of the empty seq correctly", function () { return assert.equal(0, empty().length()); });
        it("correctly drops right n items", function () { return assert.deepEqual([1, 2, 3, 4], of(1, 2, 3, 4, 5, 6).dropRight(2).toArray()); });
        it("returns an empty seq when dropping right too much", function () { return assert.deepEqual([], of(1, 2).dropRight(3).toArray()); });
        it("doesn't modify the receiver upon drop", function () {
            var x = of(1, 2, 3);
            x.drop(2);
            assert.deepEqual([1, 2, 3], x.toArray());
        });
        it("sortBy sorting works", function () { return assert.ok(of(4, 3, 2, 1)
            .equals(of(1, 2, 3, 4).sortBy(function (x, y) { return y - x; }))); });
        it("sortOn sorting works", function () { return assert.ok(of(1, 2, 3, 4)
            .equals(of(4, 2, 3, 1).sortOn(function (x) { return x; }))); });
        it("sortOn sorting works2", function () { return assert.ok(of(new SampleData_1.MyClass("zz", 1), new SampleData_1.MyClass("test", 2), new SampleData_1.MyClass("b", 3))
            .equals(of(new SampleData_1.MyClass("test", 2), new SampleData_1.MyClass("zz", 1), new SampleData_1.MyClass("b", 3)).sortOn(function (x) { return x.getField2(); }))); });
        it("sortOn sorting works with strings too", function () { return assert.ok(of(1, 2, 3, 4)
            .equals(of(4, 2, 3, 1).sortOn(function (x) { return x + ""; }))); });
        it("sortOn sorting works with strings too 2", function () { return assert.ok(of(new SampleData_1.MyClass("b", 3), new SampleData_1.MyClass("test", 2), new SampleData_1.MyClass("zz", 1))
            .equals(of(new SampleData_1.MyClass("test", 2), new SampleData_1.MyClass("zz", 1), new SampleData_1.MyClass("b", 3)).sortOn(function (x) { return x.getField1(); }))); });
        it("correctly reverses", function () { return assert.deepEqual([3, 2, 1], of(1, 2, 3).reverse().toArray()); });
        it("correctly reverses the empty vector", function () { return assert.deepEqual([], empty().reverse().toArray()); });
        it("correctly reverses also after prepend", function () { return assert.deepEqual([3, 2, 1], of(2, 3).prepend(1).reverse().toArray()); });
        it("correctly partitions also after prepend", function () { return assert.deepEqual([[1, 3, 5, 7], [2, 4, 6, 8]], of(2, 3, 4, 5, 6, 7, 8).prepend(1).partition(function (x) { return x % 2 !== 0; })
            .map(function (v) { return v.toArray(); })); });
        it("correctly infers the more precise left and right type on partition in case of typeguard", function () {
            // just checking that this compiles. 'charAt' is available on strings not numbers.
            // 'toExponential' is available on numbers not strings.
            var _a = of(1, "a")
                .partition(Comparison_1.typeOf("number")), a = _a[0], b = _a[1];
            a.single().getOrThrow().toExponential(2);
            b.single().getOrThrow().charAt(0);
        });
        it("correctly infers the more precise right type on partition in case of typeguard", function () {
            // just checking that this compiles. 'charAt' is available on strings not numbers.
            // 'toExponential' is available on numbers not strings.
            var _a = of(1, "a")
                .partition(Comparison_1.typeOf("string")), a = _a[0], b = _a[1];
            a.single().getOrThrow().charAt(0);
            b.single().getOrThrow().toExponential(2);
        });
        it("zips with an array", function () { return assert.deepEqual([[1, "a"], [2, "b"]], of(1, 2, 3).zip(["a", "b"]).toArray()); });
        it("zips with a stream", function () { return assert.deepEqual([["a", 0], ["b", 1]], of("a", "b").zip(Stream_1.Stream.iterate(0, function (x) { return x + 1; })).toArray()); });
        it("richer example", function () { return assert.deepEqual([[1, "a"], [2, "b"]], of(1, 2, 3)
            .zip(["a", "b", "c"]).takeWhile(function (_a) {
            var k = _a[0], v = _a[1];
            return k < 3;
        }).toArray()); });
        it("flatMap works", function () { return assert.ok(of(1, 2, 2, 3, 3, 3, 4, 4, 4, 4)
            .equals(of(1, 2, 3, 4).flatMap(function (x) { return ofIterable(Array.from(Array(x), function () { return x; })); }))); });
        it("map works", function () { return assert.ok(of(5, 6, 7).equals(of(1, 2, 3).map(function (x) { return x + 4; }))); });
        it("mapOption works", function () { return assert.deepEqual([3, 5, 7], of(1, 2, 3, 4, 5, 6).mapOption(function (x) { return x % 2 == 0 ? Option_1.Option.of(x + 1) : Option_1.Option.none(); }).toArray()); });
        it("supports append", function () { return assert.deepEqual([1, 2, 3, 4], of(1, 2, 3).append(4).toArray()); });
        it("doesn't modify the receiver upon append", function () {
            var x = of(1, 2, 3);
            x.append(4);
            assert.deepEqual([1, 2, 3], x.toArray());
        });
        it("supports appendAll", function () { return assert.deepEqual([1, 2, 3, 4, 5], of(1, 2, 3).appendAll([4, 5]).toArray()); });
        it("doesn't modify the receiver upon appendAll", function () {
            var x = of(1, 2, 3);
            x.appendAll([4, 5]);
            assert.deepEqual([1, 2, 3], x.toArray());
        });
        it("supports appendAll of a linkedlist iterable", function () { return assert.deepEqual([1, 2, 3, 4, 5], of(1, 2, 3).appendAll(LinkedList_1.LinkedList.of(4, 5)).toArray()); });
        it("doesn't modify the receiver upon appendAll of a linkedlist iterable", function () {
            var x = of(1, 2, 3);
            x.appendAll(LinkedList_1.LinkedList.of(4, 5));
            assert.deepEqual([1, 2, 3], x.toArray());
        });
        it("supports appendAll of a vector iterable", function () { return assert.deepEqual([1, 2, 3, 4, 5], of(1, 2, 3).appendAll(Vector_1.Vector.of(4, 5)).toArray()); });
        it("doesn't modify the receiver upon appendAll of a vector iterable", function () {
            var x = of(1, 2, 3);
            x.appendAll(Vector_1.Vector.of(4, 5));
            assert.deepEqual([1, 2, 3], x.toArray());
        });
        it("supports take", function () { return assert.deepEqual([1, 2, 3], of(1, 2, 3, 4, 5, 6).take(3).toArray()); });
        it("supports take with an index higher than the length", function () {
            return assert.deepEqual([1, 2, 3], of(1, 2, 3).take(6).toArray());
        });
        it("supports take with a negative index", function () {
            return assert.deepEqual([], of(1, 2, 3).take(-1).toArray());
        });
        it("doesn't modify the receiver upon take", function () {
            var x = of(1, 2, 3);
            x.take(2);
            assert.deepEqual([1, 2, 3], x.toArray());
        });
        check.it("applying reverse twice is the identity", gen.array(gen.string), function (ar) {
            assert.deepEqual(ofIterable(ar).reverse().reverse().toArray(), ofIterable(ar).toArray());
        });
        check.it("calling take() on the seq is the same as on the array", gen.array(gen.string), gen.int.suchThat(function (n) { return !Number.isNaN(n); }), function (ar, n) {
            assert.deepEqual(ofIterable(ar).take(n).toArray(), ar.slice(0, Math.max(0, n)));
        });
        it("correctly implements sliding on a non-empty seq", function () {
            assert.deepEqual([[1, 2], [3, 4], [5, 6]], of(1, 2, 3, 4, 5, 6).sliding(2).map(function (x) { return x.toArray(); }).toArray());
        });
        it("correctly implements sliding on a non-empty seq with length non-multiple of count", function () {
            assert.deepEqual([[1, 2, 3], [4, 5, 6], [7]], of(1, 2, 3, 4, 5, 6, 7).sliding(3).map(function (x) { return x.toArray(); }).toArray());
        });
        it("correctly implements sliding on an empty seq", function () {
            assert.deepEqual([], empty().sliding(2).map(function (x) { return x.toArray(); }).toArray());
        });
    });
    describe(seqName + " filtering", function () {
        it("filter works", function () { return assert.ok(of(2, 4)
            .equals(of(1, 2, 3, 4).filter(function (x) { return x % 2 === 0; }))); });
        it("filter works with prepend", function () { return assert.ok(of(2, 4)
            .equals(of(3, 4).prepend(2).prepend(1).filter(function (x) { return x % 2 === 0; }))); });
        it("filter upcasts if possible - instanceOf", function () {
            // here we only want to check that the code does compile,
            // that 'x' is correctly seen as MySubclass in the 3rd line
            of(new SampleData_1.MySubclass("a", 1, "b"))
                .filter(Comparison_1.instanceOf(SampleData_1.MySubclass))
                .filter(function (x) { return x.getField3().length > 1; });
        });
        it("filter upcasts if possible - typeOf", function () {
            // here we only want to check that the code does compile,
            // that 'x' is correctly seen as string in the 3rd line
            of(1, 3, "a", "b")
                .filter(Comparison_1.typeOf("string"))
                .filter(function (x) { return x.replace("x", "").length > 0; });
        });
        it("distinctBy", function () { return assert.deepEqual([1, 2, 3], of(1, 1, 2, 3, 2, 3, 1).distinctBy(function (x) { return x; }).toArray()); });
        it("distinctBy for the empty seq", function () { return assert.deepEqual([], empty().distinctBy(function (x) { return x; }).toArray()); });
        it("distinctBy for a single value", function () { return assert.deepEqual([1], of(1).distinctBy(function (x) { return x; }).toArray()); });
        it("distinctBy, custom equality", function () { return assert.deepEqual([1, 0, 2], of(1, 0, 1, 2, 3, 2, 3, 1).distinctBy(function (x) { return new SampleData_1.MyClass("hi", x % 3); }).toArray()); });
        it("distinctBy with prepend", function () { return assert.deepEqual([1, 2, 3], of(2, 3, 2, 3, 1).prepend(1).distinctBy(function (x) { return x; }).toArray()); });
        it("correctly dropsWhile", function () { return assert.deepEqual([4, 3, 5, 6], of(1, 2, 3, 4, 3, 5, 6).dropWhile(function (x) { return x < 4; }).toArray()); });
        it("correctly dropsWhile, only the first one is removed", function () { return assert.deepEqual([2, 3, 4, 3, 5, 6], of(1, 2, 3, 4, 3, 5, 6).dropWhile(function (x) { return x < 2; }).toArray()); });
        it("correctly dropsWhile, nothing matches", function () { return assert.deepEqual([], of(1, 2, 3, 4, 3, 5, 6).dropWhile(function (x) { return x >= 0; }).toArray()); });
        it("correctly dropsWhile, everything matches", function () { return assert.deepEqual([1, 2, 3, 4, 3, 5, 6], of(1, 2, 3, 4, 3, 5, 6).dropWhile(function (x) { return x < 0; }).toArray()); });
        it("correctly dropsRightWhile", function () { return assert.deepEqual([1, 4, 2, 3], of(1, 4, 2, 3, 4, 5, 6).dropRightWhile(function (x) { return x >= 4; }).toArray()); });
        it("correctly dropsRightWhile, only the last one is removed", function () { return assert.deepEqual([1, 4, 2, 3, 4, 5], of(1, 4, 2, 3, 4, 5, 6).dropRightWhile(function (x) { return x > 5; }).toArray()); });
        it("correctly dropsRightWhile, nothing matches", function () { return assert.deepEqual([], of(1, 4, 2, 3, 4, 5, 6).dropRightWhile(function (x) { return x >= 0; }).toArray()); });
        it("correctly dropsRightWhile, everything matches", function () { return assert.deepEqual([1, 4, 2, 3, 4, 5, 6], of(1, 4, 2, 3, 4, 5, 6).dropRightWhile(function (x) { return x < 0; }).toArray()); });
        it("correctly drops n items", function () { return assert.deepEqual([4, 5, 6], of(1, 2, 3, 4, 5, 6).drop(3).toArray()); });
        it("returns an empty stream when dropping too much", function () { return assert.deepEqual([], of(1, 2).drop(3).toArray()); });
        it("calculates removeFirst well", function () { return assert.deepEqual([0, 1, 3, 4], of(0, 1, 2, 3, 4).removeFirst(Predicate_1.Predicate.isIn([3, 2])).toArray()); });
        it("calculates removeFirst well event if item not present", function () { return assert.deepEqual([0, 1, 2, 3, 4], of(0, 1, 2, 3, 4).removeFirst(Predicate_1.Predicate.equals(5)).toArray()); });
        it("calculates removeFirst from empty well", function () { return assert.ok(empty().equals(empty().removeFirst(function (x) { return x === 3; }))); });
        it("handles a simple splitAt", function () { return assert.deepEqual([[1, 2, 3], [4, 5]], of(1, 2, 3, 4, 5).splitAt(3).map(function (x) { return x.toArray(); })); });
        it("handles a simple splitAt 2", function () { return assert.deepEqual([[1], [2, 3]], of(1, 2, 3).splitAt(1).map(function (x) { return x.toArray(); })); });
        it("handles splitAt with empty second list", function () { return assert.deepEqual([[1, 2, 3], []], of(1, 2, 3).splitAt(3).map(function (x) { return x.toArray(); })); });
        it("handles splitAt with empty first list", function () { return assert.deepEqual([[], [1, 2, 3]], of(1, 2, 3).splitAt(0).map(function (x) { return x.toArray(); })); });
        it("handles splitAt with out of bounds 1", function () { return assert.deepEqual([[], [1, 2, 3]], of(1, 2, 3).splitAt(-1).map(function (x) { return x.toArray(); })); });
        it("handles splitAt with out of bounds 2", function () { return assert.deepEqual([[1, 2, 3], []], of(1, 2, 3).splitAt(4).map(function (x) { return x.toArray(); })); });
        it("handles a regular span", function () { return assert.deepEqual([[1, 2], [3, 4, 5]], of(1, 2, 3, 4, 5).span(function (i) { return i < 3; }).map(function (x) { return x.toArray(); })); });
        it("handles a span matching everything", function () { return assert.deepEqual([[1, 2, 3, 4, 5], []], of(1, 2, 3, 4, 5).span(function (i) { return i < 9; }).map(function (x) { return x.toArray(); })); });
        it("handles a span matching nothing", function () { return assert.deepEqual([[], [1, 2, 3, 4, 5]], of(1, 2, 3, 4, 5).span(function (i) { return i < 0; }).map(function (x) { return x.toArray(); })); });
    });
    describe(seqName + " value extraction", function () {
        it("get finds when present", function () { return assert.ok(Option_1.Option.of(5).equals(of(1, 2, 3, 4, 5, 6).get(4))); });
        it("get finds when present after prepend", function () { return assert.ok(Option_1.Option.of(5).equals(of(2, 3, 4, 5, 6).prepend(1).get(4))); });
        it("get doesn't find when stream too short", function () { return assert.ok(Option_1.Option.none().equals(of(1, 2, 3).get(4))); });
        it("get doesn't find when negative index", function () { return assert.ok(Option_1.Option.none().equals(of(1, 2, 3).get(-1))); });
        it("correctly gets the tail of the empty vector", function () { return assert.ok(empty().tail().isNone()); });
        it("correctly gets the tail of a simple vector", function () { return assert.ok(of(2, 3, 4).equals(of(1, 2, 3, 4).tail().getOrThrow())); });
        it("correctly gets the tail of a vector after prepend", function () { return assert.ok(of(2, 3, 4).equals(of(2, 3, 4).prepend(1).tail().getOrThrow())); });
        it("gets the last value correctly", function () { return assert.equal(3, of(1, 2, 3).last().getOrThrow()); });
        it("gets the last value correctly for an empty seq", function () { return assert.ok(empty().last().isNone()); });
        it("correctly gets the last element also after prepend", function () { return assert.equal(5, of(4, 5).prependAll(of(1, 2, 3)).last().getOrUndefined()); });
        it("correctly gets the first element", function () { return assert.equal(1, of(1, 2, 3, 4, 5).head().getOrUndefined()); });
        it("correctly gets the first element of an empty vector", function () { return assert.ok(empty().head().isNone()); });
        it("correctly gets the first element also after prepend", function () { return assert.equal(1, of(4, 5).prependAll(of(1, 2, 3)).head().getOrUndefined()); });
        it("correct returns single positive case", function () { return assert.equal(5, of(5).single().getOrThrow()); });
        it("correct returns single negative case", function () { return assert.ok(of(5, 6).single().isNone()); });
        it("correct returns single empty seq", function () { return assert.ok(empty().single().isNone()); });
    });
    describe(seqName + " supports shuffle", function () {
        var original = of(1, 2, 3, 4, 5, 6, 7, 2, 9, 10);
        var shuffle1 = original.shuffle();
        var shuffle2 = original.shuffle();
        var shuffle3 = original.shuffle();
        var shuffle4 = original.shuffle();
        it("should conserve all the items1", function () { return assert.deepEqual(original.sortOn(function (x) { return x; }).toArray(), shuffle1.sortOn(function (x) { return x; }).toArray()); });
        it("should conserve all the items2", function () { return assert.deepEqual(original.sortOn(function (x) { return x; }).toArray(), shuffle2.sortOn(function (x) { return x; }).toArray()); });
        it("should conserve all the items3", function () { return assert.deepEqual(original.sortOn(function (x) { return x; }).toArray(), shuffle3.sortOn(function (x) { return x; }).toArray()); });
        it("should conserve all the items4", function () { return assert.deepEqual(original.sortOn(function (x) { return x; }).toArray(), shuffle4.sortOn(function (x) { return x; }).toArray()); });
        it("should change the order of elements", function () { return assert.equal(false, original.equals(shuffle1) &&
            original.equals(shuffle2) &&
            original.equals(shuffle3) &&
            original.equals(shuffle4)); });
    });
    var list = of({ a: 1, b: "aa" }, { a: 2, b: "aa" }, { a: 1, b: "ba" });
    describe(seqName + " multiple criteria sorting", function () {
        it("sorts normally, descending on b", function () {
            assert.deepEqual([{ a: 1, b: "ba" }, { a: 1, b: "aa" }, { a: 2, b: "aa" }], list.sortOn(function (x) { return x.a; }, { desc: function (x) { return x.b; } }).toArray());
        });
        it("sorts normally ascending on b", function () {
            assert.deepEqual([{ a: 1, b: "aa" }, { a: 1, b: "ba" }, { a: 2, b: "aa" }], list.sortOn(function (x) { return x.a; }, function (x) { return x.b; }).toArray());
        });
        it("sorts normally descending on a", function () {
            assert.deepEqual([{ a: 2, b: "aa" }, { a: 1, b: "aa" }, { a: 1, b: "ba" }], list.sortOn({ desc: function (x) { return x.a; } }, function (x) { return x.b; }).toArray());
        });
        it("sorts normally descending on both", function () {
            assert.deepEqual([{ a: 2, b: "aa" }, { a: 1, b: "ba" }, { a: 1, b: "aa" }], list.sortOn({ desc: function (x) { return x.a; } }, { desc: function (x) { return x.b; } }).toArray());
        });
    });
    var list1 = of({ a: 1, b: true }, { a: 2, b: false }, { a: 1, b: false });
    describe(seqName + " multiple criteria sorting (including booleans)", function () {
        it("sorts normally, descending on b", function () {
            assert.deepEqual([{ a: 1, b: true }, { a: 1, b: false }, { a: 2, b: false }], list1.sortOn(function (x) { return x.a; }, { desc: function (x) { return x.b; } }).toArray());
        });
        it("sorts normally ascending on b", function () {
            assert.deepEqual([{ a: 1, b: false }, { a: 2, b: false }, { a: 1, b: true }], list1.sortOn(function (x) { return x.b; }, function (x) { return x.a; }).toArray());
        });
    });
}
exports.runTests = runTests;
//# sourceMappingURL=Seq.js.map