"use strict";
exports.__esModule = true;
var Vector_1 = require("../src/Vector");
var Stream_1 = require("../src/Stream");
var SampleData_1 = require("./SampleData");
var Comparison_1 = require("../src/Comparison");
var SeqTest = require("./Seq");
var assert = require("assert");
SeqTest.runTests("Vector", Vector_1.Vector.ofIterable, Vector_1.Vector.of, Vector_1.Vector.empty, Vector_1.Vector.unfoldRight);
describe("Vector toString", function () {
    it("serializes to string correctly", function () { return assert.equal("Vector(1, 2, 3)", Vector_1.Vector.of(1, 2, 3).toString()); });
    it("serializes to string correctly - arrays & strings", function () { return assert.equal("Vector([1,'a'])", Vector_1.Vector.of([1, 'a']).toString()); });
    it("serializes to string correctly - custom toString", function () { return assert.equal("Vector({field1: hi, field2: 99})", Vector_1.Vector.of(new SampleData_1.MyClass("hi", 99)).toString()); });
    it("serializes to string correctly - plain map", function () { return assert.equal("Vector({\"name\":\"hi\",\"age\":99})", Vector_1.Vector.of({ name: "hi", age: 99 }).toString()); });
});
// tests with over 32 elements
describe("Vector over one node", function () {
    it("maps also on longer lists", function () { return assert.deepEqual([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24,
        25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39], Stream_1.Stream.iterate(1, function (x) { return x + 1; }).take(40).toVector().map(function (x) { return x - 1; }).toArray()); });
    it("iterator is correct also on longer lists", function () { return assert.deepEqual([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24,
        25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40], Array.from(Stream_1.Stream.iterate(1, function (x) { return x + 1; }).take(40).toVector())); });
    it("actually broke once", function () { return assert.ok(Stream_1.Stream.iterate(0, function (i) { return i + 1; }).take(33).toVector().equals(Stream_1.Stream.iterate(0, function (i) { return i + 1; }).take(31).toVector().appendAll([31, 32]))); });
});
describe("Vector extra methods", function () {
    it("map calls the conversion function once per element", function () {
        var i = 0;
        Vector_1.Vector.of(1, 2, 3).map(function (x) { i += 1; return x + 4; });
        assert.equal(3, i);
    });
    it("handles init correctly on a non-empty vector", function () { return assert.deepEqual([1, 2, 3], Vector_1.Vector.of(1, 2, 3, 4).init().toArray()); });
    it("handles init correctly on an empty vector", function () { return assert.deepEqual([], Vector_1.Vector.empty().init().toArray()); });
    it("handles init correctly on a single-element vector", function () { return assert.deepEqual([], Vector_1.Vector.of(1).init().toArray()); });
    it("doesn't modify the receiver on init", function () {
        var x = Vector_1.Vector.of(1, 2, 3);
        x.init();
        assert.deepEqual([1, 2, 3], x.toArray());
    });
    it("handles replace correctly on a simple example", function () { return assert.deepEqual([1, 4, 3], Vector_1.Vector.of(1, 2, 3).replace(1, 4).toArray()); });
    it("should throw on replace on negative index", function () { return assert.throws(function () { return Vector_1.Vector.of(1, 2, 3).replace(-1, 0); }); });
    it("should throw on replace on out of bounds index", function () { return assert.throws(function () { return Vector_1.Vector.of(1, 2, 3).replace(5, 0); }); });
    it("doesn't modify the receiver on replace", function () {
        var x = Vector_1.Vector.of(1, 2, 3);
        x.replace(1, 4);
        assert.deepEqual([1, 2, 3], x.toArray());
    });
    it("correctly infers the more precise left type on partition in case of typeguard", function () {
        // just checking that this compiles. 'charAt' is available on strings not numbers.
        // the get(0) is to make sure that partition returns me a Vector
        // not a Collection or something less precise than Vector.
        Vector_1.Vector.of(1, "test", 2, "a")
            .partition(Comparison_1.typeOf("string"))[0]
            .get(0).getOrThrow().charAt(0);
    });
});
// that's needed due to node's assert.deepEqual
// saying that new Array(2) is NOT the same as
// [undefined, undefined].
// (empty vs undefined)
// => forcing undefined
function arraySetUndefineds(ar) {
    if (!ar) {
        return ar;
    }
    for (var i = 0; i < ar.length; i++) {
        if (Array.isArray(ar[i])) {
            arraySetUndefineds(ar[i]);
        }
        if (typeof ar[i] === "undefined") {
            ar[i] = undefined;
        }
    }
    return ar;
}
function checkTake(longer, n, shorter) {
    var arrayBefore = longer.toArray();
    assert.deepEqual(arraySetUndefineds(shorter._contents), longer.take(n)._contents);
    // taking should not have modified the original vector
    assert.deepEqual(arrayBefore, longer.toArray());
}
// check that the internal structure of the vector trie
// is correct after take, that means also root killing and so on.
describe("Vector.take() implementation", function () {
    it("handles simple cases correctly", function () {
        return checkTake(Vector_1.Vector.of(1, 2, 3, 4, 5, 6), 3, Vector_1.Vector.of(1, 2, 3));
    });
    it("handles root killing correctly", function () { return checkTake(Vector_1.Vector.ofIterable(Stream_1.Stream.iterate(1, function (i) { return i + 1; }).take(40)), 3, Vector_1.Vector.of(1, 2, 3)); });
    it("handles double root killing correctly", function () { return checkTake(Vector_1.Vector.ofIterable(Stream_1.Stream.iterate(1, function (i) { return i + 1; }).take(1100)), 3, Vector_1.Vector.of(1, 2, 3)); });
    it("handles taking all on length multiple of node size correctly", function () { return checkTake(Stream_1.Stream.iterate(1, function (i) { return i + 1; }).take(128).toVector(), 128, Stream_1.Stream.iterate(1, function (i) { return i + 1; }).take(128).toVector()); });
    it("handles taking more than the whole length on length multiple of node size correctly", function () { return checkTake(Stream_1.Stream.iterate(1, function (i) { return i + 1; }).take(128).toVector(), 129, Stream_1.Stream.iterate(1, function (i) { return i + 1; }).take(128).toVector()); });
});
function checkAppend(base, toAppend, combined) {
    var arrayBefore = base.toArray();
    assert.deepEqual(arraySetUndefineds(combined._content), base.appendAll(toAppend)._content);
    // appending should not have modified the original vector
    assert.deepEqual(arrayBefore, base.toArray());
}
describe("Vector.appendAll() implementation", function () {
    it("handles simple cases correctly", function () {
        checkAppend(Vector_1.Vector.of(1, 2, 3), [4, 5, 6, 7, 8], Vector_1.Vector.of(1, 2, 3, 4, 5, 6, 7, 8));
    });
    it("handles adding nodes correctly", function () {
        checkAppend(Vector_1.Vector.of(1, 2, 3), Stream_1.Stream.iterate(4, function (i) { return i + 1; }).take(70), Vector_1.Vector.ofIterable(Stream_1.Stream.iterate(0, function (i) { return i + 1; }).take(74)));
    });
    it("handles adding nodes correctly, adding an array", function () {
        checkAppend(Vector_1.Vector.of(1, 2, 3), Stream_1.Stream.iterate(4, function (i) { return i + 1; }).take(70).toArray(), Vector_1.Vector.ofIterable(Stream_1.Stream.iterate(0, function (i) { return i + 1; }).take(74)));
    });
    it("handles adding nodes correctly, adding a vector", function () {
        checkAppend(Vector_1.Vector.of(1, 2, 3), Stream_1.Stream.iterate(4, function (i) { return i + 1; }).take(70).toVector(), Vector_1.Vector.ofIterable(Stream_1.Stream.iterate(0, function (i) { return i + 1; }).take(74)));
    });
});
//# sourceMappingURL=Vector.js.map